import asyncio
import json
from base64 import b64decode
from typing import List
from urllib import parse

import pytest

from gql import Client, gql

from .conftest import MS, WebSocketServerHelper

# Marking all tests in this file with the websockets marker
pytestmark = pytest.mark.websockets

SEND_MESSAGE_DELAY = 20 * MS
NB_MESSAGES = 10

DUMMY_API_KEY = "da2-thisisadummyapikey01234567"
DUMMY_ACCESS_KEY_ID = "DUMMYACCESSKEYID0123"
DUMMY_ACCESS_KEY_ID_NOT_ALLOWED = "DUMMYACCESSKEYID!ALL"
DUMMY_ACCESS_KEY_IDS = [DUMMY_ACCESS_KEY_ID, DUMMY_ACCESS_KEY_ID_NOT_ALLOWED]
DUMMY_SECRET_ACCESS_KEY = "ThisIsADummySecret0123401234012340123401"
DUMMY_SECRET_SESSION_TOKEN = (
    "FwoREDACTEDzEREDACTED+YREDACTEDJLREDACTEDz2REDACTEDH5RE"
    "DACTEDbVREDACTEDqwREDACTEDHJREDACTEDxFREDACTEDtMREDACTED5kREDACTEDSwREDACTED0BRED"
    "ACTEDuDREDACTEDm4REDACTEDSBREDACTEDaoREDACTEDP2REDACTEDCBREDACTED0wREDACTEDmdREDA"
    "CTEDyhREDACTEDSKREDACTEDYbREDACTEDfeREDACTED3UREDACTEDaKREDACTEDi1REDACTEDGEREDAC"
    "TED4VREDACTEDjmREDACTEDYcREDACTEDkQREDACTEDyI="
)
REGION_NAME = "eu-west-3"

# List which can used to store received messages by the server
logged_messages: List[str] = []


def realtime_appsync_server_factory(
    keepalive=False, not_json_answer=False, error_without_id=False
):
    def verify_headers(headers, in_query=False):
        """Returns an error or None if all is ok"""

        if "x-api-key" in headers:
            print("API KEY Authentication detected!")

            if headers["x-api-key"] == DUMMY_API_KEY:
                return None

        elif "Authorization" in headers:
            if "X-Amz-Security-Token" in headers:
                with_token = True
                print("IAM Authentication with token detected!")
            else:
                with_token = False
                print("IAM Authentication with token detected!")
                print("IAM Authentication without token detected!")

            assert headers["accept"] == "application/json, text/javascript"
            assert headers["content-encoding"] == "amz-1.0"
            assert headers["content-type"] == "application/json; charset=UTF-8"
            assert "X-Amz-Date" in headers

            authorization_fields = headers["Authorization"].split(" ")

            assert authorization_fields[0] == "AWS4-HMAC-SHA256"

            credential_field = authorization_fields[1][:-1].split("=")
            assert credential_field[0] == "Credential"
            credential_content = credential_field[1].split("/")
            assert credential_content[0] in DUMMY_ACCESS_KEY_IDS

            if in_query:
                if credential_content[0] == DUMMY_ACCESS_KEY_ID_NOT_ALLOWED:
                    return {
                        "errorType": "UnauthorizedException",
                        "message": "Permission denied",
                    }

            # assert credential_content[1]== date
            # assert credential_content[2]== region
            assert credential_content[3] == "appsync"
            assert credential_content[4] == "aws4_request"

            signed_headers_field = authorization_fields[2][:-1].split("=")

            assert signed_headers_field[0] == "SignedHeaders"
            signed_headers = signed_headers_field[1].split(";")

            assert "accept" in signed_headers
            assert "content-encoding" in signed_headers
            assert "content-type" in signed_headers
            assert "host" in signed_headers
            assert "x-amz-date" in signed_headers

            if with_token:
                assert "x-amz-security-token" in signed_headers

            signature_field = authorization_fields[3].split("=")

            assert signature_field[0] == "Signature"

            return None

        return {
            "errorType": "com.amazonaws.deepdish.graphql.auth#UnauthorizedException",
            "message": "You are not authorized to make this call.",
            "errorCode": 400,
        }

    async def realtime_appsync_server_template(ws):
        import websockets

        logged_messages.clear()

        try:
            if not_json_answer:
                await ws.send("Something not json")
                return

            if error_without_id:
                await ws.send(
                    json.dumps(
                        {
                            "type": "error",
                            "payload": {
                                "errors": [
                                    {
                                        "errorType": "Error without id",
                                        "message": (
                                            "Sometimes AppSync will send you "
                                            "an error without an id"
                                        ),
                                    }
                                ]
                            },
                        },
                        separators=(",", ":"),
                    )
                )
                return

            path = ws.request.path

            print(f"path = {path}")

            path_base, parameters_str = path.split("?")

            assert path_base == "/graphql"

            parameters = parse.parse_qs(parameters_str)

            header_param = parameters["header"][0]
            payload_param = parameters["payload"][0]

            assert payload_param == "e30="

            headers = json.loads(b64decode(header_param).decode())

            print("\nHeaders received in URL:")
            for key, value in headers.items():
                print(f"    {key}: {value}")
            print("\n")

            error = verify_headers(headers)

            if error is not None:
                await ws.send(
                    json.dumps(
                        {"payload": {"errors": [error]}, "type": "connection_error"},
                        separators=(",", ":"),
                    )
                )
                return

            await WebSocketServerHelper.send_connection_ack(
                ws, payload='{"connectionTimeoutMs":300000}'
            )

            result = await ws.recv()
            logged_messages.append(result)

            json_result = json.loads(result)

            query_id = json_result["id"]
            assert json_result["type"] == "start"

            payload = json_result["payload"]

            # With appsync, the data field is serialized to string
            data_str = payload["data"]
            extensions = payload["extensions"]

            data = json.loads(data_str)

            query = data["query"]
            variables = data.get("variables", None)
            operation_name = data.get("operationName", None)
            print(f"Received query: {query}")
            print(f"Received variables: {variables}")
            print(f"Received operation_name: {operation_name}")

            authorization = extensions["authorization"]
            print("\nHeaders received in the extensions of the query:")
            for key, value in authorization.items():
                print(f"    {key}: {value}")
            print("\n")

            error = verify_headers(headers, in_query=True)

            if error is not None:
                await ws.send(
                    json.dumps(
                        {
                            "id": str(query_id),
                            "type": "error",
                            "payload": {"errors": [error]},
                        },
                        separators=(",", ":"),
                    )
                )
                return

            await ws.send(
                json.dumps(
                    {"id": str(query_id), "type": "start_ack"}, separators=(",", ":")
                )
            )

            async def send_message_coro():
                print("            Server: send message task started")
                try:
                    for number in range(NB_MESSAGES):
                        payload = {
                            "data": {
                                "onCreateMessage": {"message": f"Hello world {number}!"}
                            }
                        }

                        if operation_name or variables:

                            payload["extensions"] = {}

                            if operation_name:
                                payload["extensions"]["operation_name"] = operation_name
                            if variables:
                                payload["extensions"]["variables"] = variables

                        await ws.send(
                            json.dumps(
                                {
                                    "id": str(query_id),
                                    "type": "data",
                                    "payload": payload,
                                },
                                separators=(",", ":"),
                            )
                        )
                        await asyncio.sleep(SEND_MESSAGE_DELAY)
                finally:
                    print("            Server: send message task ended")

            print("            Server: starting send message task")
            send_message_task = asyncio.ensure_future(send_message_coro())

            async def keepalive_coro():
                while True:
                    await asyncio.sleep(5 * MS)
                    try:
                        await WebSocketServerHelper.send_keepalive(ws)
                    except websockets.exceptions.ConnectionClosed:
                        break

            if keepalive:
                print("            Server: starting keepalive task")
                keepalive_task = asyncio.ensure_future(keepalive_coro())

            async def receiving_coro():
                print("            Server: receiving task started")
                try:
                    nonlocal send_message_task
                    while True:

                        try:
                            result = await ws.recv()
                            logged_messages.append(result)
                        except websockets.exceptions.ConnectionClosed:
                            break

                finally:
                    print("            Server: receiving task ended")
                    if keepalive:
                        keepalive_task.cancel()

            print("            Server: starting receiving task")
            receiving_task = asyncio.ensure_future(receiving_coro())

            try:
                print(
                    "            Server: waiting for sending message task to complete"
                )
                await send_message_task
            except asyncio.CancelledError:
                print("            Server: Now sending message task is cancelled")

            print("            Server: sending complete message")
            await WebSocketServerHelper.send_complete(ws, query_id)

            if keepalive:
                print("            Server: cancelling keepalive task")
                keepalive_task.cancel()
                try:
                    await keepalive_task
                except asyncio.CancelledError:
                    print("            Server: Now keepalive task is cancelled")

            print("            Server: waiting for client to close the connection")
            try:
                await asyncio.wait_for(receiving_task, 1000 * MS)
            except asyncio.TimeoutError:
                pass

            print("            Server: cancelling receiving task")
            receiving_task.cancel()

            try:
                await receiving_task
            except asyncio.CancelledError:
                print("            Server: Now receiving task is cancelled")

        except websockets.exceptions.ConnectionClosedOK:
            pass
        except AssertionError as e:
            print(f"\n            Server: Assertion failed: {e!s}\n")
        except Exception as e:
            print(f"\n            Server: Exception received: {e!s}\n")
        finally:
            print("            Server: waiting for websocket connection to close")
            try:
                await asyncio.wait_for(ws.wait_closed(), 1000 * MS)
            except asyncio.TimeoutError:
                pass

            try:
                await asyncio.wait_for(ws.close(), 1000 * MS)
            except asyncio.TimeoutError:
                pass

            print("            Server: connection closed")

    return realtime_appsync_server_template


async def realtime_appsync_server(ws):

    server = realtime_appsync_server_factory()
    await server(ws)


async def realtime_appsync_server_keepalive(ws):

    server = realtime_appsync_server_factory(keepalive=True)
    await server(ws)


async def realtime_appsync_server_not_json_answer(ws):

    server = realtime_appsync_server_factory(not_json_answer=True)
    await server(ws)


async def realtime_appsync_server_error_without_id(ws):

    server = realtime_appsync_server_factory(error_without_id=True)
    await server(ws)


on_create_message_subscription_str = """
subscription onCreateMessage {
  onCreateMessage {
    message
  }
}
"""


async def default_transport_test(transport):
    client = Client(transport=transport)

    expected_messages = [f"Hello world {number}!" for number in range(NB_MESSAGES)]
    received_messages = []

    async with client as session:
        subscription = gql(on_create_message_subscription_str)

        async for result in session.subscribe(subscription):

            message = result["onCreateMessage"]["message"]
            print(f"Message received: '{message}'")

            received_messages.append(message)

    assert expected_messages == received_messages


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [realtime_appsync_server_keepalive], indirect=True)
async def test_appsync_subscription_api_key(server):

    from gql.transport.appsync_auth import AppSyncApiKeyAuthentication
    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport

    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"

    auth = AppSyncApiKeyAuthentication(host=server.hostname, api_key=DUMMY_API_KEY)

    transport = AppSyncWebsocketsTransport(
        url=url, auth=auth, keep_alive_timeout=(5 * SEND_MESSAGE_DELAY)
    )

    await default_transport_test(transport)


@pytest.mark.asyncio
@pytest.mark.botocore
@pytest.mark.parametrize("server", [realtime_appsync_server], indirect=True)
async def test_appsync_subscription_iam_with_token(server):

    from botocore.credentials import Credentials

    from gql.transport.appsync_auth import AppSyncIAMAuthentication
    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport

    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"

    dummy_credentials = Credentials(
        access_key=DUMMY_ACCESS_KEY_ID,
        secret_key=DUMMY_SECRET_ACCESS_KEY,
        token=DUMMY_SECRET_SESSION_TOKEN,
    )

    auth = AppSyncIAMAuthentication(
        host=server.hostname, credentials=dummy_credentials, region_name=REGION_NAME
    )

    transport = AppSyncWebsocketsTransport(url=url, auth=auth)

    await default_transport_test(transport)


@pytest.mark.asyncio
@pytest.mark.botocore
@pytest.mark.parametrize("server", [realtime_appsync_server], indirect=True)
async def test_appsync_subscription_iam_without_token(server):

    from botocore.credentials import Credentials

    from gql.transport.appsync_auth import AppSyncIAMAuthentication
    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport

    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"

    dummy_credentials = Credentials(
        access_key=DUMMY_ACCESS_KEY_ID,
        secret_key=DUMMY_SECRET_ACCESS_KEY,
    )

    auth = AppSyncIAMAuthentication(
        host=server.hostname, credentials=dummy_credentials, region_name=REGION_NAME
    )

    transport = AppSyncWebsocketsTransport(url=url, auth=auth)

    await default_transport_test(transport)


@pytest.mark.asyncio
@pytest.mark.botocore
@pytest.mark.parametrize("server", [realtime_appsync_server], indirect=True)
async def test_appsync_execute_method_not_allowed(server):

    from botocore.credentials import Credentials

    from gql.transport.appsync_auth import AppSyncIAMAuthentication
    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport

    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"

    dummy_credentials = Credentials(
        access_key=DUMMY_ACCESS_KEY_ID,
        secret_key=DUMMY_SECRET_ACCESS_KEY,
    )

    auth = AppSyncIAMAuthentication(
        host=server.hostname, credentials=dummy_credentials, region_name=REGION_NAME
    )

    transport = AppSyncWebsocketsTransport(url=url, auth=auth)

    client = Client(transport=transport)

    async with client as session:
        query = gql(
            """
mutation createMessage($message: String!) {
  createMessage(input: {message: $message}) {
    id
    message
    createdAt
  }
}"""
        )

        query.variable_values = {"message": "Hello world!"}

        with pytest.raises(AssertionError) as exc_info:
            await session.execute(query)

        assert (
            "execute method is not allowed for AppSyncWebsocketsTransport "
            "because only subscriptions are allowed on the realtime endpoint."
        ) in str(exc_info)


@pytest.mark.asyncio
@pytest.mark.botocore
async def test_appsync_fetch_schema_from_transport_not_allowed():

    from botocore.credentials import Credentials

    from gql.transport.appsync_auth import AppSyncIAMAuthentication
    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport

    dummy_credentials = Credentials(
        access_key=DUMMY_ACCESS_KEY_ID,
        secret_key=DUMMY_SECRET_ACCESS_KEY,
    )

    auth = AppSyncIAMAuthentication(
        host="something", credentials=dummy_credentials, region_name=REGION_NAME
    )

    transport = AppSyncWebsocketsTransport(url="https://something", auth=auth)

    with pytest.raises(AssertionError) as exc_info:
        Client(transport=transport, fetch_schema_from_transport=True)

    assert (
        "fetch_schema_from_transport=True is not allowed for AppSyncWebsocketsTransport"
        " because only subscriptions are allowed on the realtime endpoint."
    ) in str(exc_info)


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [realtime_appsync_server], indirect=True)
async def test_appsync_subscription_api_key_unauthorized(server):

    from gql.transport.appsync_auth import AppSyncApiKeyAuthentication
    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport
    from gql.transport.exceptions import TransportServerError

    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"

    auth = AppSyncApiKeyAuthentication(host=server.hostname, api_key="invalid")

    transport = AppSyncWebsocketsTransport(url=url, auth=auth)

    client = Client(transport=transport)

    with pytest.raises(TransportServerError) as exc_info:
        async with client as _:
            pass

    assert "You are not authorized to make this call." in str(exc_info)


@pytest.mark.asyncio
@pytest.mark.botocore
@pytest.mark.parametrize("server", [realtime_appsync_server], indirect=True)
async def test_appsync_subscription_iam_not_allowed(server):

    from botocore.credentials import Credentials

    from gql.transport.appsync_auth import AppSyncIAMAuthentication
    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport
    from gql.transport.exceptions import TransportQueryError

    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"

    dummy_credentials = Credentials(
        access_key=DUMMY_ACCESS_KEY_ID_NOT_ALLOWED,
        secret_key=DUMMY_SECRET_ACCESS_KEY,
        token=DUMMY_SECRET_SESSION_TOKEN,
    )

    auth = AppSyncIAMAuthentication(
        host=server.hostname, credentials=dummy_credentials, region_name=REGION_NAME
    )

    transport = AppSyncWebsocketsTransport(url=url, auth=auth)

    client = Client(transport=transport)

    async with client as session:
        subscription = gql(on_create_message_subscription_str)

        with pytest.raises(TransportQueryError) as exc_info:

            async for result in session.subscribe(subscription):
                pass

        assert "Permission denied" in str(exc_info)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "server", [realtime_appsync_server_not_json_answer], indirect=True
)
async def test_appsync_subscription_server_sending_a_not_json_answer(server):

    from gql.transport.appsync_auth import AppSyncApiKeyAuthentication
    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport
    from gql.transport.exceptions import TransportProtocolError

    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"

    auth = AppSyncApiKeyAuthentication(host=server.hostname, api_key=DUMMY_API_KEY)

    transport = AppSyncWebsocketsTransport(url=url, auth=auth)

    client = Client(transport=transport)

    with pytest.raises(TransportProtocolError) as exc_info:
        async with client as _:
            pass

    assert "Server did not return a GraphQL result: Something not json" in str(exc_info)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "server", [realtime_appsync_server_error_without_id], indirect=True
)
async def test_appsync_subscription_server_sending_an_error_without_an_id(server):

    from gql.transport.appsync_auth import AppSyncApiKeyAuthentication
    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport
    from gql.transport.exceptions import TransportServerError

    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"

    auth = AppSyncApiKeyAuthentication(host=server.hostname, api_key=DUMMY_API_KEY)

    transport = AppSyncWebsocketsTransport(url=url, auth=auth)

    client = Client(transport=transport)

    with pytest.raises(TransportServerError) as exc_info:
        async with client as _:
            pass

    assert "Sometimes AppSync will send you an error without an id" in str(exc_info)


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [realtime_appsync_server_keepalive], indirect=True)
async def test_appsync_subscription_variable_values_and_operation_name(server):

    from gql.transport.appsync_auth import AppSyncApiKeyAuthentication
    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport

    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"

    auth = AppSyncApiKeyAuthentication(host=server.hostname, api_key=DUMMY_API_KEY)

    transport = AppSyncWebsocketsTransport(
        url=url, auth=auth, keep_alive_timeout=(5 * SEND_MESSAGE_DELAY)
    )

    client = Client(transport=transport)

    expected_messages = [f"Hello world {number}!" for number in range(NB_MESSAGES)]
    received_messages = []

    async with client as session:
        subscription = gql(on_create_message_subscription_str)

        subscription.variable_values = {"key1": "val1"}
        subscription.operation_name = "onCreateMessage"

        async for execution_result in session.subscribe(
            subscription,
            get_execution_result=True,
        ):

            result = execution_result.data
            message = result["onCreateMessage"]["message"]
            print(f"Message received: '{message}'")

            received_messages.append(message)

            print(f"extensions received: {execution_result.extensions}")

            assert execution_result.extensions["operation_name"] == "onCreateMessage"
            variables = execution_result.extensions["variables"]
            assert variables["key1"] == "val1"

    assert expected_messages == received_messages



import pytest
from graphql import print_schema

from gql import Client

from .fixtures import make_starwars_transport

# Marking all tests in this file with the aiohttp marker
pytestmark = pytest.mark.aiohttp


@pytest.mark.asyncio
async def test_starwars_introspection_args(aiohttp_server):

    transport = await make_starwars_transport(aiohttp_server)

    # First fetch the schema from transport using default introspection query
    # We should receive descriptions in the schema but not deprecated input fields
    async with Client(
        transport=transport,
        fetch_schema_from_transport=True,
        introspection_args={
            "input_value_deprecation": False,
        },
    ) as session:

        schema_str = print_schema(session.client.schema)
        print(schema_str)

        assert '"""The number of stars this review gave, 1-5"""' in schema_str
        assert "deprecated_input_field" not in schema_str

    # Then fetch the schema from transport using an introspection query
    # without requesting descriptions
    # We should NOT receive descriptions in the schema
    async with Client(
        transport=transport,
        fetch_schema_from_transport=True,
        introspection_args={
            "descriptions": False,
            "input_value_deprecation": False,
        },
    ) as session:

        schema_str = print_schema(session.client.schema)
        print(schema_str)

        assert '"""The number of stars this review gave, 1-5"""' not in schema_str
        assert "deprecated_input_field" not in schema_str

    # Then fetch the schema from transport using and introspection query
    # requiring deprecated input fields
    # We should receive descriptions in the schema and deprecated input fields
    async with Client(
        transport=transport,
        fetch_schema_from_transport=True,
    ) as session:

        schema_str = print_schema(session.client.schema)
        print(schema_str)

        assert '"""The number of stars this review gave, 1-5"""' in schema_str
        assert "deprecated_input_field" in schema_str



import asyncio
import os
import sys
from urllib.parse import urlparse

from gql import Client, gql
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.appsync_auth import AppSyncIAMAuthentication

# Uncomment the following lines to enable debug output
# import logging
# logging.basicConfig(level=logging.DEBUG)


async def main():

    # Should look like:
    # https://XXXXXXXXXXXXXXXXXXXXXXXXXX.appsync-api.REGION.amazonaws.com/graphql
    url = os.environ.get("AWS_GRAPHQL_API_ENDPOINT")

    if url is None:
        print("Missing environment variables")
        sys.exit()

    # Extract host from url
    host = str(urlparse(url).netloc)

    auth = AppSyncIAMAuthentication(host=host)

    transport = AIOHTTPTransport(url=url, auth=auth)

    async with Client(
        transport=transport,
        fetch_schema_from_transport=False,
    ) as session:

        query = gql(
            """
mutation createMessage($message: String!) {
  createMessage(input: {message: $message}) {
    id
    message
    createdAt
  }
}"""
        )

        query.variable_values = {"message": "Hello world!"}

        result = await session.execute(query)
        print(result)


asyncio.run(main())



import copy

import pytest

from gql import Client, gql

from .schema import StarWarsIntrospection, StarWarsSchema


@pytest.fixture
def local_schema():
    return Client(schema=StarWarsSchema)


@pytest.fixture
def typedef_schema():
    return Client(
        schema="""
schema {
  query: Query
}

interface Character {
  appearsIn: [Episode]
  friends: [Character]
  id: String!
  name: String
}

type Droid implements Character {
  appearsIn: [Episode]
  friends: [Character]
  id: String!
  name: String
  primaryFunction: String
}

enum Episode {
  EMPIRE
  JEDI
  NEWHOPE
}

type Human implements Character {
  appearsIn: [Episode]
  friends: [Character]
  homePlanet: String
  id: String!
  name: String
}

type Query {
  droid(id: String!): Droid
  hero(episode: Episode): Character
  human(id: String!): Human
}"""
    )


@pytest.fixture
def introspection_schema():
    return Client(introspection=StarWarsIntrospection)


@pytest.fixture
def introspection_schema_empty_directives():
    # Create a deep copy to avoid modifying the original
    introspection = copy.deepcopy(StarWarsIntrospection)

    # Simulate an empty dictionary for directives
    introspection["__schema"]["directives"] = []

    return Client(introspection=introspection)


@pytest.fixture
def introspection_schema_no_directives():
    # Create a deep copy to avoid modifying the original
    introspection = copy.deepcopy(StarWarsIntrospection)

    # Simulate no directives key
    del introspection["__schema"]["directives"]  # type: ignore

    return Client(introspection=introspection)


@pytest.fixture(
    params=[
        "local_schema",
        "typedef_schema",
        "introspection_schema",
        "introspection_schema_empty_directives",
        "introspection_schema_no_directives",
    ]
)
def client(request):
    return request.getfixturevalue(request.param)


def validation_errors(client, query):
    query = gql(query)
    try:
        client.validate(query)
        return False
    except Exception:
        return True


def test_incompatible_request_gql(client):
    with pytest.raises(TypeError):
        gql(123)  # type: ignore

    """
    The error generated depends on graphql-core version
    < 3.1.5: "body must be a string"
    >= 3.1.5: some variation of "object of type 'int' has no len()"
              depending on the python environment

    So we are not going to check the exact error message here anymore.
    """

    """
    assert ("body must be a string" in str(exc_info.value)) or (
        "object of type 'int' has no len()" in str(exc_info.value)
    )
    """


def test_nested_query_with_fragment(client):
    query = """
        query NestedQueryWithFragment {
          hero {
            ...NameAndAppearances
            friends {
              ...NameAndAppearances
              friends {
                ...NameAndAppearances
              }
            }
          }
        }
        fragment NameAndAppearances on Character {
          name
          appearsIn
        }
    """
    assert not validation_errors(client, query)


def test_non_existent_fields(client):
    query = """
        query HeroSpaceshipQuery {
          hero {
            favoriteSpaceship
          }
        }
    """
    assert validation_errors(client, query)


def test_require_fields_on_object(client):
    query = """
        query HeroNoFieldsQuery {
          hero
        }
    """
    assert validation_errors(client, query)


def test_disallows_fields_on_scalars(client):
    query = """
        query HeroFieldsOnScalarQuery {
          hero {
            name {
              firstCharacterOfName
            }
          }
        }
    """
    assert validation_errors(client, query)


def test_disallows_object_fields_on_interfaces(client):
    query = """
        query DroidFieldOnCharacter {
          hero {
            name
            primaryFunction
          }
        }
    """
    assert validation_errors(client, query)


def test_allows_object_fields_in_fragments(client):
    query = """
        query DroidFieldInFragment {
          hero {
            name
            ...DroidFields
          }
        }
        fragment DroidFields on Droid {
          primaryFunction
        }
    """
    assert not validation_errors(client, query)


def test_allows_object_fields_in_inline_fragments(client):
    query = """
        query DroidFieldInFragment {
          hero {
            name
            ... on Droid {
              primaryFunction
            }
          }
        }
    """
    assert not validation_errors(client, query)


def test_include_directive(client):
    query = """
        query fetchHero($with_friends: Boolean!) {
          hero {
            name
            friends @include(if: $with_friends) {
                name
            }
          }
        }
    """
    assert not validation_errors(client, query)


def test_skip_directive(client):
    query = """
        query fetchHero($without_friends: Boolean!) {
          hero {
            name
            friends @skip(if: $without_friends) {
                name
            }
          }
        }
    """
    assert not validation_errors(client, query)


def test_build_client_schema_invalid_introspection():
    from gql.utilities import build_client_schema

    with pytest.raises(TypeError) as exc_info:
        build_client_schema("blah")  # type: ignore

    assert (
        "Invalid or incomplete introspection result. Ensure that you are passing the "
        "'data' attribute of an introspection response and no 'errors' were returned "
        "alongside: 'blah'."
    ) in str(exc_info.value)



# First install fastapi and uvicorn:
#
# pip install fastapi uvicorn
#
# then run:
#
# uvicorn fastapi_async:app --reload

import logging

from fastapi import FastAPI, HTTPException
from fastapi.responses import HTMLResponse

from gql import Client, gql
from gql.client import ReconnectingAsyncClientSession
from gql.transport.aiohttp import AIOHTTPTransport

logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)

transport = AIOHTTPTransport(url="https://countries.trevorblades.com/graphql")

client = Client(transport=transport)

query = gql(
    """
query getContinentInfo($code: ID!) {
  continent(code:$code) {
    name
    code
    countries  {
      name
      capital
    }
  }
}
"""
)

app = FastAPI()


@app.on_event("startup")
async def startup_event():
    print("Connecting to GraphQL backend")

    await client.connect_async(reconnecting=True)
    print("End of startup")


@app.on_event("shutdown")
async def shutdown_event():
    print("Shutting down GraphQL permanent connection...")
    await client.close_async()
    print("Shutting down GraphQL permanent connection... done")


continent_codes = [
    "AF",
    "AN",
    "AS",
    "EU",
    "NA",
    "OC",
    "SA",
]


@app.get("/", response_class=HTMLResponse)
def get_root():

    continent_links = ", ".join(
        [f'<a href="continent/{code}">{code}</a>' for code in continent_codes]
    )

    return f"""
    <html>
        <head>
            <title>Continents</title>
        </head>
        <body>
            Continents: {continent_links}
        </body>
    </html>
"""


@app.get("/continent/{continent_code}")
async def get_continent(continent_code):

    if continent_code not in continent_codes:
        raise HTTPException(status_code=404, detail="Continent not found")

    try:
        assert isinstance(client.session, ReconnectingAsyncClientSession)
        query.variable_values = {"code": continent_code}
        result = await client.session.execute(query)
    except Exception as e:
        log.debug(f"get_continent Error: {e}")
        raise HTTPException(status_code=503, detail="GraphQL backend unavailable")

    return result



import pytest

mock_transport_host = "appsyncapp.awsgateway.com.example.org"
mock_transport_url = f"https://{mock_transport_host}/graphql"


@pytest.mark.botocore
def test_appsync_init_with_minimal_args(fake_session_factory):
    from gql.transport.appsync_auth import AppSyncIAMAuthentication
    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport

    transport = AppSyncWebsocketsTransport(
        url=mock_transport_url, session=fake_session_factory()
    )
    assert isinstance(transport.auth, AppSyncIAMAuthentication)
    assert transport.connect_timeout == 10
    assert transport.close_timeout == 10
    assert transport.ack_timeout == 10
    assert transport.ssl is False
    assert transport.connect_args == {}


@pytest.mark.botocore
def test_appsync_init_with_no_credentials(caplog, fake_session_factory):
    import botocore.exceptions

    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport

    with pytest.raises(botocore.exceptions.NoCredentialsError):
        transport = AppSyncWebsocketsTransport(
            url=mock_transport_url,
            session=fake_session_factory(credentials=None),
        )
        assert transport.auth is None

    expected_error = "Credentials not found"

    print(f"Captured log: {caplog.text}")

    assert expected_error in caplog.text


@pytest.mark.websockets
def test_appsync_init_with_jwt_auth():
    from gql.transport.appsync_auth import AppSyncJWTAuthentication
    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport

    auth = AppSyncJWTAuthentication(host=mock_transport_host, jwt="some-jwt")
    transport = AppSyncWebsocketsTransport(url=mock_transport_url, auth=auth)
    assert transport.auth is auth

    assert auth.get_headers() == {
        "host": mock_transport_host,
        "Authorization": "some-jwt",
    }


@pytest.mark.websockets
def test_appsync_init_with_apikey_auth():
    from gql.transport.appsync_auth import AppSyncApiKeyAuthentication
    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport

    auth = AppSyncApiKeyAuthentication(host=mock_transport_host, api_key="some-api-key")
    transport = AppSyncWebsocketsTransport(url=mock_transport_url, auth=auth)
    assert transport.auth is auth

    assert auth.get_headers() == {
        "host": mock_transport_host,
        "x-api-key": "some-api-key",
    }


@pytest.mark.botocore
def test_appsync_init_with_iam_auth_without_creds(fake_session_factory):
    import botocore.exceptions

    from gql.transport.appsync_auth import AppSyncIAMAuthentication
    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport

    auth = AppSyncIAMAuthentication(
        host=mock_transport_host,
        session=fake_session_factory(credentials=None),
    )
    with pytest.raises(botocore.exceptions.NoCredentialsError):
        AppSyncWebsocketsTransport(url=mock_transport_url, auth=auth)


@pytest.mark.botocore
def test_appsync_init_with_iam_auth_with_creds(fake_credentials_factory):
    from gql.transport.appsync_auth import AppSyncIAMAuthentication
    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport

    auth = AppSyncIAMAuthentication(
        host=mock_transport_host,
        credentials=fake_credentials_factory(),
        region_name="us-east-1",
    )
    transport = AppSyncWebsocketsTransport(url=mock_transport_url, auth=auth)
    assert transport.auth is auth


@pytest.mark.botocore
def test_appsync_init_with_iam_auth_and_no_region(
    caplog, fake_credentials_factory, fake_session_factory
):
    """

    WARNING: this test will fail if:
     - you have a default region set in ~/.aws/config
     - you have the AWS_DEFAULT_REGION environment variable set

    """
    import logging

    from botocore.exceptions import NoRegionError

    from gql.transport.appsync_auth import AppSyncIAMAuthentication
    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport

    caplog.set_level(logging.WARNING)

    with pytest.raises(NoRegionError):
        session = fake_session_factory(credentials=fake_credentials_factory())
        session._region_name = None
        session._credentials.region = None
        transport = AppSyncWebsocketsTransport(url=mock_transport_url, session=session)

        assert isinstance(transport.auth, AppSyncIAMAuthentication)

        # prints the region name in case the test fails
        print(f"Region found: {transport.auth._region_name}")

    print(f"Captured: {caplog.text}")

    expected_error = (
        "Region name not found. "
        "It was not possible to detect your region either from the host "
        "or from your default AWS configuration."
    )

    assert expected_error in caplog.text


@pytest.mark.botocore
def test_munge_url(fake_signer_factory, fake_request_factory):
    from gql.transport.appsync_auth import AppSyncIAMAuthentication
    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport

    test_url = "https://appsync-api.aws.example.org/some-other-params"

    auth = AppSyncIAMAuthentication(
        host=test_url,
        signer=fake_signer_factory(),
        request_creator=fake_request_factory,
    )
    transport = AppSyncWebsocketsTransport(url=test_url, auth=auth)

    header_string = (
        "eyJGYWtlQXV0aG9yaXphdGlvbiI6ImEiLCJGYWtlVGltZSI6InRvZGF5"
        "IiwiaG9zdCI6Imh0dHBzOi8vYXBwc3luYy1hcGkuYXdzLmV4YW1wbGUu"
        "b3JnL3NvbWUtb3RoZXItcGFyYW1zIn0="
    )
    expected_url = (
        "wss://appsync-realtime-api.aws.example.org/"
        f"some-other-params?header={header_string}&payload=e30="
    )
    assert transport.url == expected_url


@pytest.mark.botocore
def test_munge_url_format(
    fake_signer_factory,
    fake_request_factory,
    fake_credentials_factory,
    fake_session_factory,
):
    from gql.transport.appsync_auth import AppSyncIAMAuthentication

    test_url = "https://appsync-api.aws.example.org/some-other-params"

    auth = AppSyncIAMAuthentication(
        host=test_url,
        signer=fake_signer_factory(),
        session=fake_session_factory(),
        request_creator=fake_request_factory,
        credentials=fake_credentials_factory(),
    )

    header_string = (
        "eyJGYWtlQXV0aG9yaXphdGlvbiI6ImEiLCJGYWtlVGltZSI6InRvZGF5"
        "IiwiaG9zdCI6Imh0dHBzOi8vYXBwc3luYy1hcGkuYXdzLmV4YW1wbGUu"
        "b3JnL3NvbWUtb3RoZXItcGFyYW1zIn0="
    )
    expected_url = (
        "wss://appsync-realtime-api.aws.example.org/"
        f"some-other-params?header={header_string}&payload=e30="
    )
    assert auth.get_auth_url(test_url) == expected_url



"""The primary :mod:`gql` package includes everything you need to
execute GraphQL requests, with the exception of the transports
which are optional:

 - the :func:`gql <gql.gql>` method to parse a GraphQL query
 - the :class:`Client <gql.Client>` class as the entrypoint to execute requests
   and create sessions
"""

from .__version__ import __version__
from .client import Client
from .gql import gql
from .graphql_request import GraphQLRequest
from .transport.file_upload import FileVar

__all__ = [
    "__version__",
    "gql",
    "Client",
    "GraphQLRequest",
    "FileVar",
]



from graphql import DirectiveLocation, GraphQLSchema, IntrospectionQuery
from graphql import build_client_schema as build_client_schema_orig
from graphql.pyutils import inspect
from graphql.utilities.get_introspection_query import IntrospectionDirective

__all__ = ["build_client_schema"]


INCLUDE_DIRECTIVE_JSON: IntrospectionDirective = {
    "name": "include",
    "description": (
        "Directs the executor to include this field or fragment "
        "only when the `if` argument is true."
    ),
    "locations": [
        DirectiveLocation.FIELD,
        DirectiveLocation.FRAGMENT_SPREAD,
        DirectiveLocation.INLINE_FRAGMENT,
    ],
    "args": [
        {
            "name": "if",
            "description": "Included when true.",
            "type": {
                "kind": "NON_NULL",
                "name": "None",
                "ofType": {"kind": "SCALAR", "name": "Boolean", "ofType": "None"},
            },
            "defaultValue": "None",
        }
    ],
}

SKIP_DIRECTIVE_JSON: IntrospectionDirective = {
    "name": "skip",
    "description": (
        "Directs the executor to skip this field or fragment "
        "when the `if` argument is true."
    ),
    "locations": [
        DirectiveLocation.FIELD,
        DirectiveLocation.FRAGMENT_SPREAD,
        DirectiveLocation.INLINE_FRAGMENT,
    ],
    "args": [
        {
            "name": "if",
            "description": "Skipped when true.",
            "type": {
                "kind": "NON_NULL",
                "name": "None",
                "ofType": {"kind": "SCALAR", "name": "Boolean", "ofType": "None"},
            },
            "defaultValue": "None",
        }
    ],
}


def build_client_schema(introspection: IntrospectionQuery) -> GraphQLSchema:
    """This is an alternative to the graphql-core function
    :code:`build_client_schema` but with default include and skip directives
    added to the schema to fix
    `issue #278 <https://github.com/graphql-python/gql/issues/278>`_

    .. warning::
        This function will be removed once the issue
        `graphql-js#3419 <https://github.com/graphql/graphql-js/issues/3419>`_
        has been fixed and ported to graphql-core so don't use it
        outside gql.
    """

    if not isinstance(introspection, dict) or not isinstance(
        introspection.get("__schema"), dict
    ):
        raise TypeError(
            "Invalid or incomplete introspection result. Ensure that you"
            " are passing the 'data' attribute of an introspection response"
            f" and no 'errors' were returned alongside: {inspect(introspection)}."
        )

    schema_introspection = introspection["__schema"]

    directives = schema_introspection.get("directives", None)

    if directives is None:
        schema_introspection["directives"] = directives = []

    if not any(directive["name"] == "skip" for directive in directives):
        directives.append(SKIP_DIRECTIVE_JSON)

    if not any(directive["name"] == "include" for directive in directives):
        directives.append(INCLUDE_DIRECTIVE_JSON)

    return build_client_schema_orig(introspection, assume_valid=False)



import pytest


class FakeSession:
    def __init__(self, credentials, region_name):
        self._credentials = credentials
        self._region_name = region_name

    def get_default_client_config(self):
        return

    def get_credentials(self):
        return self._credentials

    def _resolve_region_name(self, region_name, client_config):
        return region_name if region_name else self._region_name


@pytest.fixture
def fake_session_factory(fake_credentials_factory):
    def _fake_session_factory(credentials=fake_credentials_factory()):
        return FakeSession(credentials=credentials, region_name="fake-region")

    yield _fake_session_factory



from typing import Mapping

import pytest

from gql import Client, GraphQLRequest, gql
from gql.transport.exceptions import (
    TransportClosed,
    TransportProtocolError,
    TransportQueryError,
    TransportServerError,
)

# Marking all tests in this file with the requests marker
pytestmark = pytest.mark.requests

query1_str = """
    query getContinents {
      continents {
        code
        name
      }
    }
"""

query1_server_answer_list = (
    '[{"data":{"continents":['
    '{"code":"AF","name":"Africa"},{"code":"AN","name":"Antarctica"},'
    '{"code":"AS","name":"Asia"},{"code":"EU","name":"Europe"},'
    '{"code":"NA","name":"North America"},{"code":"OC","name":"Oceania"},'
    '{"code":"SA","name":"South America"}]}}]'
)

query1_server_answer_twice_list = (
    "["
    '{"data":{"continents":['
    '{"code":"AF","name":"Africa"},{"code":"AN","name":"Antarctica"},'
    '{"code":"AS","name":"Asia"},{"code":"EU","name":"Europe"},'
    '{"code":"NA","name":"North America"},{"code":"OC","name":"Oceania"},'
    '{"code":"SA","name":"South America"}]}},'
    '{"data":{"continents":['
    '{"code":"AF","name":"Africa"},{"code":"AN","name":"Antarctica"},'
    '{"code":"AS","name":"Asia"},{"code":"EU","name":"Europe"},'
    '{"code":"NA","name":"North America"},{"code":"OC","name":"Oceania"},'
    '{"code":"SA","name":"South America"}]}}'
    "]"
)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_requests_query(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.requests import RequestsHTTPTransport

    async def handler(request):
        return web.Response(
            text=query1_server_answer_list,
            content_type="application/json",
            headers={"dummy": "test1234"},
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    def test_code():
        transport = RequestsHTTPTransport(url=url)

        with Client(transport=transport) as session:

            query = [GraphQLRequest(query1_str)]

            # Execute query synchronously
            results = session.execute_batch(query)

            continents = results[0]["continents"]

            africa = continents[0]

            assert africa["code"] == "AF"

            # Checking response headers are saved in the transport
            assert hasattr(transport, "response_headers")
            assert isinstance(transport.response_headers, Mapping)
            assert transport.response_headers["dummy"] == "test1234"

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_requests_query_auto_batch_enabled(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.requests import RequestsHTTPTransport

    async def handler(request):
        return web.Response(
            text=query1_server_answer_list,
            content_type="application/json",
            headers={"dummy": "test1234"},
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    def test_code():
        transport = RequestsHTTPTransport(url=url)

        with Client(
            transport=transport,
            batch_interval=0.01,
        ) as session:

            query = gql(query1_str)

            # Execute query synchronously
            result = session.execute(query)

            continents = result["continents"]

            africa = continents[0]

            assert africa["code"] == "AF"

            # Checking response headers are saved in the transport
            assert hasattr(transport, "response_headers")
            assert isinstance(transport.response_headers, Mapping)
            assert transport.response_headers["dummy"] == "test1234"

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_requests_query_auto_batch_enabled_two_requests(
    aiohttp_server, run_sync_test
):
    from threading import Thread

    from aiohttp import web

    from gql.transport.requests import RequestsHTTPTransport

    async def handler(request):
        return web.Response(
            text=query1_server_answer_twice_list,
            content_type="application/json",
            headers={"dummy": "test1234"},
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    def test_code():
        transport = RequestsHTTPTransport(url=url)

        threads = []

        with Client(
            transport=transport,
            batch_interval=0.01,
        ) as session:

            def test_thread():
                query = gql(query1_str)

                # Execute query synchronously
                result = session.execute(query)

                continents = result["continents"]

                africa = continents[0]

                assert africa["code"] == "AF"

                # Checking response headers are saved in the transport
                assert hasattr(transport, "response_headers")
                assert isinstance(transport.response_headers, Mapping)
                assert transport.response_headers["dummy"] == "test1234"

            for _ in range(2):
                thread = Thread(target=test_thread)
                thread.start()
                threads.append(thread)

        for thread in threads:
            thread.join()

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_requests_cookies(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.requests import RequestsHTTPTransport

    async def handler(request):
        assert "COOKIE" in request.headers
        assert "cookie1=val1" == request.headers["COOKIE"]

        return web.Response(
            text=query1_server_answer_list, content_type="application/json"
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    def test_code():
        transport = RequestsHTTPTransport(url=url, cookies={"cookie1": "val1"})

        with Client(transport=transport) as session:

            query = [GraphQLRequest(query1_str)]

            # Execute query synchronously
            results = session.execute_batch(query)

            continents = results[0]["continents"]

            africa = continents[0]

            assert africa["code"] == "AF"

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_requests_error_code_401(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.requests import RequestsHTTPTransport

    async def handler(request):
        # Will generate http error code 401
        return web.Response(
            text='{"error":"Unauthorized","message":"401 Client Error: Unauthorized"}',
            content_type="application/json",
            status=401,
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    def test_code():
        transport = RequestsHTTPTransport(url=url)

        with Client(transport=transport) as session:

            query = [GraphQLRequest(query1_str)]

            with pytest.raises(TransportServerError) as exc_info:
                session.execute_batch(query)

            assert "401 Client Error: Unauthorized" in str(exc_info.value)

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_requests_error_code_401_auto_batch_enabled(
    aiohttp_server, run_sync_test
):
    from aiohttp import web

    from gql.transport.requests import RequestsHTTPTransport

    async def handler(request):
        # Will generate http error code 401
        return web.Response(
            text='{"error":"Unauthorized","message":"401 Client Error: Unauthorized"}',
            content_type="application/json",
            status=401,
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    def test_code():
        transport = RequestsHTTPTransport(url=url)

        with Client(
            transport=transport,
            batch_interval=0.01,
        ) as session:

            query = gql(query1_str)

            with pytest.raises(TransportServerError) as exc_info:
                session.execute(query)

            assert "401 Client Error: Unauthorized" in str(exc_info.value)

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_requests_error_code_429(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.requests import RequestsHTTPTransport

    async def handler(request):
        # Will generate http error code 429
        return web.Response(
            text="""
<html>
  <head>
     <title>Too Many Requests</title>
  </head>
  <body>
     <h1>Too Many Requests</h1>
     <p>I only allow 50 requests per hour to this Web site per
        logged in user.  Try again soon.</p>
  </body>
</html>""",
            content_type="text/html",
            status=429,
            headers={"Retry-After": "3600"},
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    def test_code():
        transport = RequestsHTTPTransport(url=url)

        with Client(transport=transport) as session:

            query = [GraphQLRequest(query1_str)]

            with pytest.raises(TransportServerError) as exc_info:
                session.execute_batch(query)

            assert "429, message='Too Many Requests'" in str(exc_info.value)

        # Checking response headers are saved in the transport
        assert hasattr(transport, "response_headers")
        assert isinstance(transport.response_headers, Mapping)
        assert transport.response_headers["Retry-After"] == "3600"


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_requests_error_code_500(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.requests import RequestsHTTPTransport

    async def handler(request):
        # Will generate http error code 500
        raise Exception("Server error")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    def test_code():
        transport = RequestsHTTPTransport(url=url)

        with Client(transport=transport) as session:

            query = [GraphQLRequest(query1_str)]

            with pytest.raises(TransportServerError):
                session.execute_batch(query)

    await run_sync_test(server, test_code)


query1_server_error_answer_list = '[{"errors": ["Error 1", "Error 2"]}]'


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_requests_error_code(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.requests import RequestsHTTPTransport

    async def handler(request):
        return web.Response(
            text=query1_server_error_answer_list, content_type="application/json"
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    def test_code():
        transport = RequestsHTTPTransport(url=url)

        with Client(transport=transport) as session:

            query = [GraphQLRequest(query1_str)]

            with pytest.raises(TransportQueryError):
                session.execute_batch(query)

    await run_sync_test(server, test_code)


invalid_protocol_responses = [
    "{}",
    "qlsjfqsdlkj",
    '{"not_data_or_errors": 35}',
    "[{}]",
    "[qlsjfqsdlkj]",
    '[{"not_data_or_errors": 35}]',
    "[]",
    "[1]",
]


@pytest.mark.aiohttp
@pytest.mark.asyncio
@pytest.mark.parametrize("response", invalid_protocol_responses)
async def test_requests_invalid_protocol(aiohttp_server, response, run_sync_test):
    from aiohttp import web

    from gql.transport.requests import RequestsHTTPTransport

    async def handler(request):
        return web.Response(text=response, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    def test_code():
        transport = RequestsHTTPTransport(url=url)

        with Client(transport=transport) as session:

            query = [GraphQLRequest(query1_str)]

            with pytest.raises(TransportProtocolError):
                session.execute_batch(query)

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_requests_cannot_execute_if_not_connected(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.requests import RequestsHTTPTransport

    async def handler(request):
        return web.Response(
            text=query1_server_answer_list, content_type="application/json"
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    def test_code():
        transport = RequestsHTTPTransport(url=url)

        query = [GraphQLRequest(query1_str)]

        with pytest.raises(TransportClosed):
            transport.execute_batch(query)

    await run_sync_test(server, test_code)


query1_server_answer_with_extensions_list = (
    '[{"data":{"continents":['
    '{"code":"AF","name":"Africa"},{"code":"AN","name":"Antarctica"},'
    '{"code":"AS","name":"Asia"},{"code":"EU","name":"Europe"},'
    '{"code":"NA","name":"North America"},{"code":"OC","name":"Oceania"},'
    '{"code":"SA","name":"South America"}]},'
    '"extensions": {"key1": "val1"}'
    "}]"
)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_requests_query_with_extensions(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.requests import RequestsHTTPTransport

    async def handler(request):
        return web.Response(
            text=query1_server_answer_with_extensions_list,
            content_type="application/json",
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    def test_code():
        transport = RequestsHTTPTransport(url=url)

        with Client(transport=transport) as session:

            query = [GraphQLRequest(query1_str)]

            execution_results = session.execute_batch(query, get_execution_result=True)

            assert execution_results[0].extensions["key1"] == "val1"

    await run_sync_test(server, test_code)


ONLINE_URL = "https://countries.trevorblades.workers.dev/graphql"


@pytest.mark.online
@pytest.mark.requests
def test_requests_sync_batch_auto():

    from threading import Thread

    from gql.transport.requests import RequestsHTTPTransport

    client = Client(
        transport=RequestsHTTPTransport(url=ONLINE_URL),
        batch_interval=0.01,
        batch_max=3,
    )

    query = gql(
        """
        query getContinentName($continent_code: ID!) {
          continent(code: $continent_code) {
            name
          }
        }
        """
    )

    def get_continent_name(session, continent_code):
        variables = {
            "continent_code": continent_code,
        }

        result = session.execute(query, variable_values=variables)

        name = result["continent"]["name"]
        print(f"The continent with the code {continent_code} has the name: '{name}'")

    continent_codes = ["EU", "AF", "NA", "OC", "SA", "AS", "AN"]

    with client as session:

        for continent_code in continent_codes:

            thread = Thread(
                target=get_continent_name,
                args=(
                    session,
                    continent_code,
                ),
            )
            thread.start()
            thread.join()

    # Doing it twice to check that everything is closing and reconnecting correctly
    with client as session:

        for continent_code in continent_codes:

            thread = Thread(
                target=get_continent_name,
                args=(
                    session,
                    continent_code,
                ),
            )
            thread.start()
            thread.join()


@pytest.mark.online
@pytest.mark.requests
def test_requests_sync_batch_auto_execute_future():

    from gql.transport.requests import RequestsHTTPTransport

    client = Client(
        transport=RequestsHTTPTransport(url=ONLINE_URL),
        batch_interval=0.01,
        batch_max=3,
    )

    query = """
        query getContinentName($continent_code: ID!) {
          continent(code: $continent_code) {
            name
          }
        }
    """

    with client as session:

        request_eu = GraphQLRequest(query, variable_values={"continent_code": "EU"})
        future_result_eu = session._execute_future(request_eu)

        request_af = GraphQLRequest(query, variable_values={"continent_code": "AF"})
        future_result_af = session._execute_future(request_af)

        result_eu = future_result_eu.result().data
        result_af = future_result_af.result().data

        assert result_eu["continent"]["name"] == "Europe"
        assert result_af["continent"]["name"] == "Africa"


@pytest.mark.online
@pytest.mark.requests
def test_requests_sync_batch_manual():

    from gql.transport.requests import RequestsHTTPTransport

    client = Client(
        transport=RequestsHTTPTransport(url=ONLINE_URL),
    )

    query = """
        query getContinentName($continent_code: ID!) {
          continent(code: $continent_code) {
            name
          }
        }
    """

    with client as session:

        request_eu = GraphQLRequest(query, variable_values={"continent_code": "EU"})
        request_af = GraphQLRequest(query, variable_values={"continent_code": "AF"})

        result_eu, result_af = session.execute_batch([request_eu, request_af])

        assert result_eu["continent"]["name"] == "Europe"
        assert result_af["continent"]["name"] == "Africa"



import json

from graphql import (
    GraphQLArgument,
    GraphQLField,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInt,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLString,
)

nestedInput = GraphQLInputObjectType(
    "Nested",
    description="The input object that has a field pointing to itself",
    fields={"foo": GraphQLInputField(GraphQLInt, description="foo")},
)

nestedInput.fields["child"] = GraphQLInputField(nestedInput, description="child")

queryType = GraphQLObjectType(
    "Query",
    fields=lambda: {
        "echo": GraphQLField(
            args={"nested": GraphQLArgument(type_=nestedInput)},
            resolve=lambda *args, **kwargs: json.dumps(kwargs["nested"]),
            type_=GraphQLString,
        ),
    },
)

NestedInputSchema = GraphQLSchema(
    query=queryType,
    types=[nestedInput],
)



from ssl import SSLContext
from typing import Any, Dict, List, Optional, Union

from websockets.datastructures import HeadersLike

from .common.adapters.websockets import WebSocketsAdapter
from .websockets_protocol import WebsocketsProtocolTransportBase


class WebsocketsTransport(WebsocketsProtocolTransportBase):
    """:ref:`Async Transport <async_transports>` used to execute GraphQL queries on
    remote servers with websocket connection.

    This transport uses asyncio and the websockets library in order to send requests
    on a websocket connection.
    """

    def __init__(
        self,
        url: str,
        *,
        headers: Optional[HeadersLike] = None,
        ssl: Union[SSLContext, bool] = False,
        init_payload: Optional[Dict[str, Any]] = None,
        connect_timeout: Optional[Union[int, float]] = 10,
        close_timeout: Optional[Union[int, float]] = 10,
        ack_timeout: Optional[Union[int, float]] = 10,
        keep_alive_timeout: Optional[Union[int, float]] = None,
        ping_interval: Optional[Union[int, float]] = None,
        pong_timeout: Optional[Union[int, float]] = None,
        answer_pings: bool = True,
        connect_args: Optional[Dict[str, Any]] = None,
        subprotocols: Optional[List[str]] = None,
    ) -> None:
        """Initialize the transport with the given parameters.

        :param url: The GraphQL server URL. Example: 'wss://server.com:PORT/graphql'.
        :param headers: Dict of HTTP Headers.
        :param ssl: ssl_context of the connection. Use ssl=False to disable encryption
        :param init_payload: Dict of the payload sent in the connection_init message.
        :param connect_timeout: Timeout in seconds for the establishment
            of the websocket connection. If None is provided this will wait forever.
        :param close_timeout: Timeout in seconds for the close. If None is provided
            this will wait forever.
        :param ack_timeout: Timeout in seconds to wait for the connection_ack message
            from the server. If None is provided this will wait forever.
        :param keep_alive_timeout: Optional Timeout in seconds to receive
            a sign of liveness from the server.
        :param ping_interval: Delay in seconds between pings sent by the client to
            the backend for the graphql-ws protocol. None (by default) means that
            we don't send pings. Note: there are also pings sent by the underlying
            websockets protocol. See the
            :ref:`keepalive documentation <websockets_transport_keepalives>`
            for more information about this.
        :param pong_timeout: Delay in seconds to receive a pong from the backend
            after we sent a ping (only for the graphql-ws protocol).
            By default equal to half of the ping_interval.
        :param answer_pings: Whether the client answers the pings from the backend
            (for the graphql-ws protocol).
            By default: True
        :param connect_args: Other parameters forwarded to
            `websockets.connect <https://websockets.readthedocs.io/en/stable/reference/\
            client.html#opening-a-connection>`_
        :param subprotocols: list of subprotocols sent to the
            backend in the 'subprotocols' http header.
            By default: both apollo and graphql-ws subprotocols.
        """

        # Instanciate a WebSocketAdapter to indicate the use
        # of the websockets dependency for this transport
        self.adapter: WebSocketsAdapter = WebSocketsAdapter(
            url=url,
            headers=headers,
            ssl=ssl,
            connect_args=connect_args,
        )

        # Initialize the WebsocketsProtocolTransportBase parent class
        super().__init__(
            adapter=self.adapter,
            init_payload=init_payload,
            connect_timeout=connect_timeout,
            close_timeout=close_timeout,
            ack_timeout=ack_timeout,
            keep_alive_timeout=keep_alive_timeout,
            ping_interval=ping_interval,
            pong_timeout=pong_timeout,
            answer_pings=answer_pings,
            subprotocols=subprotocols,
        )

    @property
    def headers(self) -> Optional[HeadersLike]:
        return self.adapter.headers

    @property
    def ssl(self) -> Union[SSLContext, bool]:
        return self.adapter.ssl



from typing import (
    Any,
    Dict,
    List,
)

from graphql import ExecutionResult

from ...graphql_request import GraphQLRequest
from ..exceptions import (
    TransportProtocolError,
)


def _raise_protocol_error(result_text: str, reason: str) -> None:
    raise TransportProtocolError(
        f"Server did not return a valid GraphQL result: " f"{reason}: " f"{result_text}"
    )


def _validate_answer_is_a_list(results: Any) -> None:
    if not isinstance(results, list):
        _raise_protocol_error(
            str(results),
            "Answer is not a list",
        )


def _validate_data_and_errors_keys_in_answers(results: List[Dict[str, Any]]) -> None:
    for result in results:
        if "errors" not in result and "data" not in result:
            _raise_protocol_error(
                str(results),
                'No "data" or "errors" keys in answer',
            )


def _validate_every_answer_is_a_dict(results: List[Dict[str, Any]]) -> None:
    for result in results:
        if not isinstance(result, dict):
            _raise_protocol_error(str(results), "Not every answer is dict")


def _validate_num_of_answers_same_as_requests(
    reqs: List[GraphQLRequest],
    results: List[Dict[str, Any]],
) -> None:
    if len(reqs) != len(results):
        _raise_protocol_error(
            str(results),
            (
                "Invalid number of answers: "
                f"{len(results)} answers received for {len(reqs)} requests"
            ),
        )


def _answer_to_execution_result(result: Dict[str, Any]) -> ExecutionResult:
    return ExecutionResult(
        errors=result.get("errors"),
        data=result.get("data"),
        extensions=result.get("extensions"),
    )


def get_batch_execution_result_list(
    reqs: List[GraphQLRequest],
    answers: List,
) -> List[ExecutionResult]:

    _validate_answer_is_a_list(answers)
    _validate_num_of_answers_same_as_requests(reqs, answers)
    _validate_every_answer_is_a_dict(answers)
    _validate_data_and_errors_keys_in_answers(answers)

    return [_answer_to_execution_result(answer) for answer in answers]



import os

from setuptools import setup, find_packages

install_requires = [
    "graphql-core>=3.3.0a3,<3.4",
    "yarl>=1.6,<2.0",
    "backoff>=1.11.1,<3.0",
    "anyio>=3.0,<5",
]

console_scripts = [
    "gql-cli=gql.cli:gql_cli",
]

tests_requires = [
    "parse==1.20.2",
    "pytest==8.3.4",
    "pytest-asyncio==0.25.3",
    "pytest-console-scripts==1.4.1",
    "pytest-cov==6.0.0",
    "vcrpy==7.0.0",
    "aiofiles",
]

dev_requires = [
    "black==25.1.0",
    "check-manifest>=0.42,<1",
    "flake8==7.1.2",
    "isort==6.0.1",
    "mypy==1.15",
    "sphinx>=7.0.0,<8;python_version<='3.9'",
    "sphinx>=8.1.0,<9;python_version>'3.9'",
    "sphinx_rtd_theme>=3.0.2,<4",
    "sphinx-argparse==0.5.2",
    "types-aiofiles",
    "types-requests",
] + tests_requires

install_aiohttp_requires = [
    "aiohttp>=3.11.2,<4",
]

install_requests_requires = [
    "requests>=2.26,<3",
    "requests_toolbelt>=1.0.0,<2",
]

install_httpx_requires = [
    "httpx>=0.27.0,<1",
]

install_websockets_requires = [
    "websockets>=14.2,<16",
]

install_botocore_requires = [
    "botocore>=1.21,<2",
]

install_aiofiles_requires = [
    "aiofiles",
]

install_all_requires = (
    install_aiohttp_requires + install_requests_requires + install_httpx_requires + install_websockets_requires + install_botocore_requires + install_aiofiles_requires
)

# Get version from __version__.py file
current_folder = os.path.abspath(os.path.dirname(__file__))
about = {}
with open(os.path.join(current_folder, "gql", "__version__.py")) as f:
    exec(f.read(), about)

setup(
    name="gql",
    version=about["__version__"],
    description="GraphQL client for Python",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/graphql-python/gql",
    author="Syrus Akbary",
    author_email="me@syrusakbary.com",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    keywords="api graphql protocol rest relay gql client",
    packages=find_packages(include=["gql*"]),
    # PEP-561: https://www.python.org/dev/peps/pep-0561/
    package_data={"gql": ["py.typed"]},
    install_requires=install_requires,
    extras_require={
        "all": install_all_requires,
        "test": install_all_requires + tests_requires,
        "test_no_transport": tests_requires,
        "dev": install_all_requires + dev_requires,
        "aiohttp": install_aiohttp_requires,
        "requests": install_requests_requires,
        "httpx": install_httpx_requires,
        "websockets": install_websockets_requires,
        "botocore": install_botocore_requires,
        "aiofiles": install_aiofiles_requires,
    },
    include_package_data=True,
    zip_safe=False,
    platforms="any",
    entry_points={"console_scripts": console_scripts},
)



from .graphql_request import GraphQLRequest


def gql(request_string: str) -> GraphQLRequest:
    """Given a string containing a GraphQL request,
       parse it into a Document and put it into a GraphQLRequest object.

    :param request_string: the GraphQL request as a String
    :return: a :class:`GraphQLRequest <gql.GraphQLRequest>`
             which can be later executed or subscribed by a
             :class:`Client <gql.client.Client>`, by an
             :class:`async session <gql.client.AsyncClientSession>` or by a
             :class:`sync session <gql.client.SyncClientSession>`
    :raises graphql.error.GraphQLError: if a syntax error is encountered.
    """
    return GraphQLRequest(request_string)



import asyncio
from typing import List

import pytest

from gql import Client, gql
from gql.transport.exceptions import (
    TransportConnectionFailed,
    TransportProtocolError,
    TransportQueryError,
)

from .conftest import WebSocketServerHelper

# Marking all tests in this file with the websockets marker
pytestmark = pytest.mark.websockets

invalid_query_str = """
    query getContinents {
      continents {
        code
        bloh
      }
    }
"""

invalid_query1_server_answer = (
    '{{"type":"next","id":"{query_id}",'
    '"payload":{{"errors":['
    '{{"message":"Cannot query field \\"bloh\\" on type \\"Continent\\".",'
    '"locations":[{{"line":4,"column":5}}],'
    '"extensions":{{"code":"INTERNAL_SERVER_ERROR"}}}}]}}}}'
)

invalid_query1_server = [invalid_query1_server_answer]


@pytest.mark.asyncio
@pytest.mark.parametrize("graphqlws_server", [invalid_query1_server], indirect=True)
@pytest.mark.parametrize("query_str", [invalid_query_str])
async def test_graphqlws_invalid_query(client_and_graphqlws_server, query_str):

    session, server = client_and_graphqlws_server

    query = gql(query_str)

    with pytest.raises(TransportQueryError) as exc_info:
        await session.execute(query)

    exception = exc_info.value

    assert isinstance(exception.errors, List)

    error = exception.errors[0]

    assert error["extensions"]["code"] == "INTERNAL_SERVER_ERROR"


invalid_subscription_str = """
    subscription getContinents {
      continents {
        code
        bloh
      }
    }
"""


async def server_invalid_subscription(ws):
    await WebSocketServerHelper.send_connection_ack(ws)
    await ws.recv()
    await ws.send(invalid_query1_server_answer.format(query_id=1))
    await WebSocketServerHelper.send_complete(ws, 1)
    await ws.wait_closed()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "graphqlws_server", [server_invalid_subscription], indirect=True
)
@pytest.mark.parametrize("query_str", [invalid_subscription_str])
async def test_graphqlws_invalid_subscription(client_and_graphqlws_server, query_str):

    session, server = client_and_graphqlws_server

    query = gql(query_str)

    with pytest.raises(TransportQueryError) as exc_info:
        async for result in session.subscribe(query):
            pass

    exception = exc_info.value

    assert isinstance(exception.errors, List)

    error = exception.errors[0]

    assert error["extensions"]["code"] == "INTERNAL_SERVER_ERROR"


async def server_no_ack(ws):
    await ws.wait_closed()


@pytest.mark.asyncio
@pytest.mark.parametrize("graphqlws_server", [server_no_ack], indirect=True)
@pytest.mark.parametrize("query_str", [invalid_query_str])
async def test_graphqlws_server_does_not_send_ack(graphqlws_server, query_str):
    from gql.transport.websockets import WebsocketsTransport

    url = f"ws://{graphqlws_server.hostname}:{graphqlws_server.port}/graphql"

    transport = WebsocketsTransport(url=url, ack_timeout=0.1)

    with pytest.raises(asyncio.TimeoutError):
        async with Client(transport=transport):
            pass


invalid_query_server_answer = (
    '{"id":"1","type":"error","payload":[{"message":"Cannot query field '
    '\\"helo\\" on type \\"Query\\". Did you mean \\"hello\\"?",'
    '"locations":[{"line":2,"column":3}]}]}'
)


async def server_invalid_query(ws):
    await WebSocketServerHelper.send_connection_ack(ws)
    result = await ws.recv()
    print(f"Server received: {result}")
    await ws.send(invalid_query_server_answer)
    await WebSocketServerHelper.wait_connection_terminate(ws)
    await ws.wait_closed()


@pytest.mark.asyncio
@pytest.mark.parametrize("graphqlws_server", [server_invalid_query], indirect=True)
async def test_graphqlws_sending_invalid_query(client_and_graphqlws_server):

    session, server = client_and_graphqlws_server

    query = gql("{helo}")

    with pytest.raises(TransportQueryError) as exc_info:
        await session.execute(query)

    exception = exc_info.value

    assert isinstance(exception.errors, List)

    error = exception.errors[0]

    assert (
        error["message"]
        == 'Cannot query field "helo" on type "Query". Did you mean "hello"?'
    )


not_json_answer = ["BLAHBLAH"]
missing_type_answer = ["{}"]
missing_id_answer_1 = ['{"type": "next"}']
missing_id_answer_2 = ['{"type": "error"}']
missing_id_answer_3 = ['{"type": "complete"}']
data_without_payload = ['{"type": "next", "id":"1"}']
error_without_payload = ['{"type": "error", "id":"1"}']
error_with_payload_not_a_list = ['{"type": "error", "id":"1", "payload": "NOT A LIST"}']
payload_is_not_a_dict = ['{"type": "next", "id":"1", "payload": "BLAH"}']
empty_payload = ['{"type": "next", "id":"1", "payload": {}}']
sending_bytes = [b"\x01\x02\x03"]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "graphqlws_server",
    [
        not_json_answer,
        missing_type_answer,
        missing_id_answer_1,
        missing_id_answer_2,
        missing_id_answer_3,
        data_without_payload,
        error_without_payload,
        payload_is_not_a_dict,
        error_with_payload_not_a_list,
        empty_payload,
        sending_bytes,
    ],
    indirect=True,
)
async def test_graphqlws_transport_protocol_errors(client_and_graphqlws_server):

    session, server = client_and_graphqlws_server

    query = gql("query { hello }")

    with pytest.raises(TransportProtocolError):
        await session.execute(query)


async def server_without_ack(ws):
    # Sending something else than an ack
    await WebSocketServerHelper.send_complete(ws, 1)
    await ws.wait_closed()


@pytest.mark.asyncio
@pytest.mark.parametrize("graphqlws_server", [server_without_ack], indirect=True)
async def test_graphqlws_server_does_not_ack(graphqlws_server):
    from gql.transport.websockets import WebsocketsTransport

    url = f"ws://{graphqlws_server.hostname}:{graphqlws_server.port}/graphql"
    print(f"url = {url}")

    transport = WebsocketsTransport(url=url)

    with pytest.raises(TransportProtocolError):
        async with Client(transport=transport):
            pass


async def server_closing_directly(ws):
    await ws.close()


@pytest.mark.asyncio
@pytest.mark.parametrize("graphqlws_server", [server_closing_directly], indirect=True)
async def test_graphqlws_server_closing_directly(graphqlws_server):
    from gql.transport.websockets import WebsocketsTransport

    url = f"ws://{graphqlws_server.hostname}:{graphqlws_server.port}/graphql"
    print(f"url = {url}")

    transport = WebsocketsTransport(url=url)

    with pytest.raises(TransportConnectionFailed):
        async with Client(transport=transport):
            pass


async def server_closing_after_ack(ws):
    await WebSocketServerHelper.send_connection_ack(ws)
    await ws.close()


@pytest.mark.asyncio
@pytest.mark.parametrize("graphqlws_server", [server_closing_after_ack], indirect=True)
async def test_graphqlws_server_closing_after_ack(client_and_graphqlws_server):

    session, server = client_and_graphqlws_server

    query = gql("query { hello }")

    print("\n Trying to execute first query.\n")

    with pytest.raises(TransportConnectionFailed) as exc1:
        await session.execute(query)

    exc1_cause = exc1.value.__cause__
    exc1_cause_str = f"{type(exc1_cause).__name__}:{exc1_cause!s}"

    print(f"\n First query Exception cause: {exc1_cause_str}\n")

    assert (
        exc1_cause_str == "ConnectionClosedOK:received 1000 (OK); then sent 1000 (OK)"
    )

    await session.transport.wait_closed()

    print("\n Trying to execute second query.\n")

    with pytest.raises(TransportConnectionFailed) as exc2:
        await session.execute(query)

    exc2_cause = exc2.value.__cause__
    exc2_cause_str = f"{type(exc2_cause).__name__}:{exc2_cause!s}"

    print(f" Second query Exception cause: {exc2_cause_str}\n")

    assert (
        exc2_cause_str == "ConnectionClosedOK:received 1000 (OK); then sent 1000 (OK)"
    )



import asyncio
import logging

from gql import Client, gql
from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

logging.basicConfig(level=logging.INFO)


async def main():

    transport = AIOHTTPWebsocketsTransport(
        url="wss://countries.trevorblades.com/graphql"
    )

    # Using `async with` on the client will start a connection on the transport
    # and provide a `session` variable to execute queries on this connection
    async with Client(
        transport=transport,
    ) as session:

        # Execute single query
        query = gql(
            """
            query getContinents {
              continents {
                code
                name
              }
            }
        """
        )
        result = await session.execute(query)
        print(result)

        # Request subscription
        subscription = gql(
            """
            subscription {
                somethingChanged {
                    id
                }
            }
        """
        )
        async for result in session.subscribe(subscription):
            print(result)


asyncio.run(main())



import asyncio
from typing import Collection


class Character:
    id: str
    name: str
    friends: Collection[str]
    appearsIn: Collection[str]


# noinspection PyPep8Naming
class Human(Character):
    type = "Human"
    homePlanet: str

    # noinspection PyShadowingBuiltins
    def __init__(self, id, name, friends, appearsIn, homePlanet):
        self.id, self.name = id, name
        self.friends, self.appearsIn = friends, appearsIn
        self.homePlanet = homePlanet


# noinspection PyPep8Naming
class Droid(Character):
    type = "Droid"
    primaryFunction: str

    # noinspection PyShadowingBuiltins
    def __init__(self, id, name, friends, appearsIn, primaryFunction):
        self.id, self.name = id, name
        self.friends, self.appearsIn = friends, appearsIn
        self.primaryFunction = primaryFunction


luke = Human(
    id="1000",
    name="Luke Skywalker",
    friends=["1002", "1003", "2000", "2001"],
    appearsIn=[4, 5, 6],
    homePlanet="Tatooine",
)

vader = Human(
    id="1001",
    name="Darth Vader",
    friends=["1004"],
    appearsIn=[4, 5, 6],
    homePlanet="Tatooine",
)

han = Human(
    id="1002",
    name="Han Solo",
    friends=["1000", "1003", "2001"],
    appearsIn=[4, 5, 6],
    homePlanet=None,
)

leia = Human(
    id="1003",
    name="Leia Organa",
    friends=["1000", "1002", "2000", "2001"],
    appearsIn=[4, 5, 6],
    homePlanet="Alderaan",
)

tarkin = Human(
    id="1004",
    name="Wilhuff Tarkin",
    friends=["1001"],
    appearsIn=[4],
    homePlanet=None,
)

humanData = {
    "1000": luke,
    "1001": vader,
    "1002": han,
    "1003": leia,
    "1004": tarkin,
}

threepio = Droid(
    id="2000",
    name="C-3PO",
    friends=["1000", "1002", "1003", "2001"],
    appearsIn=[4, 5, 6],
    primaryFunction="Protocol",
)

artoo = Droid(
    id="2001",
    name="R2-D2",
    friends=["1000", "1002", "1003"],
    appearsIn=[4, 5, 6],
    primaryFunction="Astromech",
)

droidData = {
    "2000": threepio,
    "2001": artoo,
}

reviews = {
    4: [{"stars": 4, "commentary": "Was good.", "episode": 4}],
    5: [{"stars": 5, "commentary": "This is a great movie!", "episode": 5}],
    6: [{"stars": 3, "commentary": "Was expecting more stuff", "episode": 6}],
}


def get_character(id):
    return humanData.get(id) or droidData.get(id)


def get_characters(ids):
    return map(get_character, ids)


def get_friends(character):
    return map(get_character, character.friends)


def get_hero(episode):
    if episode == 5:
        return luke
    return artoo


async def get_hero_async(episode):
    await asyncio.sleep(0.001)
    return get_hero(episode)


def get_human(id):
    return humanData.get(id)


def get_droid(id):
    return droidData.get(id)


def create_review(episode, review):
    reviews[episode].append(review)
    review["episode"] = episode
    return review


async def make_starwars_backend(aiohttp_server):
    from aiohttp import web
    from graphql import graphql_sync

    from .schema import StarWarsSchema

    async def handler(request):
        data = await request.json()
        source = data["query"]

        try:
            variables = data["variables"]
        except KeyError:
            variables = None

        result = graphql_sync(StarWarsSchema, source, variable_values=variables)

        return web.json_response(
            {
                "data": result.data,
                "errors": [str(e) for e in result.errors] if result.errors else None,
            }
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    return server


async def make_starwars_transport(aiohttp_server):
    from gql.transport.aiohttp import AIOHTTPTransport

    server = await make_starwars_backend(aiohttp_server)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    return transport



import io
import json
import os
import warnings
from typing import Mapping

import pytest

from gql import Client, FileVar, gql
from gql.cli import get_parser, main
from gql.transport.exceptions import (
    TransportAlreadyConnected,
    TransportClosed,
    TransportConnectionFailed,
    TransportProtocolError,
    TransportQueryError,
    TransportServerError,
)

from .conftest import (
    TemporaryFile,
    get_localhost_ssl_context_client,
    make_upload_handler,
)

query1_str = """
    query getContinents {
      continents {
        code
        name
      }
    }
"""

query1_server_answer_data = (
    '{"continents":['
    '{"code":"AF","name":"Africa"},{"code":"AN","name":"Antarctica"},'
    '{"code":"AS","name":"Asia"},{"code":"EU","name":"Europe"},'
    '{"code":"NA","name":"North America"},{"code":"OC","name":"Oceania"},'
    '{"code":"SA","name":"South America"}]}'
)


query1_server_answer = f'{{"data":{query1_server_answer_data}}}'

# Marking all tests in this file with the aiohttp marker
pytestmark = pytest.mark.aiohttp


@pytest.mark.asyncio
async def test_aiohttp_query(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(
            text=query1_server_answer,
            content_type="application/json",
            headers={"dummy": "test1234"},
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    async with Client(transport=transport) as session:

        query = gql(query1_str)

        # Execute query asynchronously
        result = await session.execute(query)

        continents = result["continents"]

        africa = continents[0]

        assert africa["code"] == "AF"

        # Checking response headers are saved in the transport
        assert hasattr(transport, "response_headers")
        assert isinstance(transport.response_headers, Mapping)
        assert transport.response_headers["dummy"] == "test1234"


@pytest.mark.asyncio
async def test_aiohttp_ignore_backend_content_type(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(text=query1_server_answer, content_type="text/plain")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    async with Client(transport=transport) as session:

        query = gql(query1_str)

        result = await session.execute(query)

        continents = result["continents"]

        africa = continents[0]

        assert africa["code"] == "AF"


@pytest.mark.asyncio
async def test_aiohttp_cookies(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        assert "COOKIE" in request.headers
        assert "cookie1=val1" == request.headers["COOKIE"]

        return web.Response(text=query1_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, cookies={"cookie1": "val1"})

    async with Client(transport=transport) as session:

        query = gql(query1_str)

        # Execute query asynchronously
        result = await session.execute(query)

        continents = result["continents"]

        africa = continents[0]

        assert africa["code"] == "AF"


@pytest.mark.asyncio
async def test_aiohttp_error_code_401(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        # Will generate http error code 401
        return web.Response(
            text='{"error":"Unauthorized","message":"401 Client Error: Unauthorized"}',
            content_type="application/json",
            status=401,
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url)

    async with Client(transport=transport) as session:

        query = gql(query1_str)

        with pytest.raises(TransportServerError) as exc_info:
            await session.execute(query)

        assert "401, message='Unauthorized'" in str(exc_info.value)


@pytest.mark.asyncio
async def test_aiohttp_error_code_429(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        # Will generate http error code 429
        return web.Response(
            text="""
<html>
  <head>
     <title>Too Many Requests</title>
  </head>
  <body>
     <h1>Too Many Requests</h1>
     <p>I only allow 50 requests per hour to this Web site per
        logged in user.  Try again soon.</p>
  </body>
</html>""",
            content_type="text/html",
            status=429,
            headers={"Retry-After": "3600"},
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url)

    async with Client(transport=transport) as session:

        query = gql(query1_str)

        with pytest.raises(TransportServerError) as exc_info:
            await session.execute(query)

        assert "429, message='Too Many Requests'" in str(exc_info.value)

        # Checking response headers are saved in the transport
        assert hasattr(transport, "response_headers")
        assert isinstance(transport.response_headers, Mapping)
        assert transport.response_headers["Retry-After"] == "3600"


@pytest.mark.asyncio
async def test_aiohttp_error_code_500(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        # Will generate http error code 500
        raise Exception("Server error")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url)

    async with Client(transport=transport) as session:

        query = gql(query1_str)

        with pytest.raises(TransportServerError) as exc_info:
            await session.execute(query)

        assert "500, message='Internal Server Error'" in str(exc_info.value)


transport_query_error_responses = [
    '{"errors": ["Error 1", "Error 2"]}',
    '{"errors": {"error_1": "Something"}}',
    '{"errors": 5}',
]


@pytest.mark.asyncio
@pytest.mark.parametrize("query_error", transport_query_error_responses)
async def test_aiohttp_error_code(aiohttp_server, query_error):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(text=query_error, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url)

    async with Client(transport=transport) as session:

        query = gql(query1_str)

        with pytest.raises(TransportQueryError):
            await session.execute(query)


invalid_protocol_responses = [
    {
        "response": "{}",
        "expected_exception": (
            "Server did not return a valid GraphQL result: "
            'No "data" or "errors" keys in answer: {}'
        ),
    },
    {
        "response": "qlsjfqsdlkj",
        "expected_exception": (
            "Server did not return a valid GraphQL result: "
            "Not a JSON answer: qlsjfqsdlkj"
        ),
    },
    {
        "response": '{"not_data_or_errors": 35}',
        "expected_exception": (
            "Server did not return a valid GraphQL result: "
            'No "data" or "errors" keys in answer: {"not_data_or_errors": 35}'
        ),
    },
    {
        "response": "",
        "expected_exception": (
            "Server did not return a valid GraphQL result: Not a JSON answer: "
        ),
    },
]


@pytest.mark.asyncio
@pytest.mark.parametrize("param", invalid_protocol_responses)
async def test_aiohttp_invalid_protocol(aiohttp_server, param):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    response = param["response"]

    async def handler(request):
        return web.Response(text=response, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url)

    async with Client(transport=transport) as session:

        query = gql(query1_str)

        with pytest.raises(TransportProtocolError) as exc_info:
            await session.execute(query)

        assert param["expected_exception"] in str(exc_info.value)


@pytest.mark.asyncio
async def test_aiohttp_subscribe_not_supported(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(text="does not matter", content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url)

    async with Client(transport=transport) as session:

        query = gql(query1_str)

        with pytest.raises(NotImplementedError):
            async for result in session.subscribe(query):
                pass


@pytest.mark.asyncio
async def test_aiohttp_cannot_connect_twice(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(text=query1_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    async with Client(transport=transport) as session:

        with pytest.raises(TransportAlreadyConnected):
            await session.transport.connect()


@pytest.mark.asyncio
async def test_aiohttp_cannot_execute_if_not_connected(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(text=query1_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    query = gql(query1_str)

    with pytest.raises(TransportClosed):
        await transport.execute(query)


@pytest.mark.asyncio
async def test_aiohttp_extra_args(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(text=query1_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    # passing extra arguments to aiohttp.ClientSession
    from aiohttp import DummyCookieJar

    jar = DummyCookieJar()
    transport = AIOHTTPTransport(
        url=url, timeout=10, client_session_args={"version": "1.1", "cookie_jar": jar}
    )

    async with Client(transport=transport) as session:

        query = gql(query1_str)

        # Passing extra arguments to the post method of aiohttp
        result = await session.execute(query, extra_args={"allow_redirects": False})

        continents = result["continents"]

        africa = continents[0]

        assert africa["code"] == "AF"


query2_str = """
    query getEurope ($code: ID!) {
      continent (code: $code) {
        name
      }
    }
"""

query2_server_answer = '{"data": {"continent": {"name": "Europe"}}}'


@pytest.mark.asyncio
async def test_aiohttp_query_variable_values(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(text=query2_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    async with Client(transport=transport) as session:

        query = gql(query2_str)

        query.variable_values = {"code": "EU"}
        query.operation_name = "getEurope"

        # Execute query asynchronously
        result = await session.execute(query)

        continent = result["continent"]

        assert continent["name"] == "Europe"


@pytest.mark.asyncio
async def test_aiohttp_query_variable_values_fix_issue_292(aiohttp_server):
    """Allow to specify variable_values without keyword.

    See https://github.com/graphql-python/gql/issues/292"""

    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(text=query2_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    async with Client(transport=transport) as session:

        query = gql(query2_str)

        query.variable_values = {"code": "EU"}
        query.operation_name = "getEurope"

        # Execute query asynchronously
        result = await session.execute(query)

        continent = result["continent"]

        assert continent["name"] == "Europe"


@pytest.mark.asyncio
async def test_aiohttp_execute_running_in_thread(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(text=query1_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    def test_code():
        transport = AIOHTTPTransport(url=url)

        client = Client(transport=transport)

        query = gql(query1_str)

        client.execute(query)

    await run_sync_test(server, test_code)


@pytest.mark.asyncio
async def test_aiohttp_subscribe_running_in_thread(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(text=query1_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    def test_code():
        transport = AIOHTTPTransport(url=url)

        client = Client(transport=transport)

        query = gql(query1_str)

        # Note: subscriptions are not supported on the aiohttp transport
        # But we add this test in order to have 100% code coverage
        # It is to check that we will correctly set an event loop
        # in the subscribe function if there is none (in a Thread for example)
        # We cannot test this with the websockets transport because
        # the websockets transport will set an event loop in its init

        with pytest.raises(NotImplementedError):
            for result in client.subscribe(query):
                pass

    await run_sync_test(server, test_code)


file_upload_mutation_1 = """
    mutation($file: Upload!) {
      uploadFile(input:{other_var:$other_var, file:$file}) {
        success
      }
    }
"""

file_upload_mutation_1_operations = (
    '{"query": "mutation ($file: Upload!) {\\n  uploadFile(input: {other_var: '
    '$other_var, file: $file}) {\\n    success\\n  }\\n}", "variables": '
    '{"file": null, "other_var": 42}}'
)

file_upload_mutation_1_map = '{"0": ["variables.file"]}'

file_1_content = """
This is a test file
This file will be sent in the GraphQL mutation
"""


@pytest.mark.asyncio
async def test_aiohttp_file_upload(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    app = web.Application()
    app.router.add_route(
        "POST",
        "/",
        make_upload_handler(
            expected_map=file_upload_mutation_1_map,
            expected_operations=file_upload_mutation_1_operations,
            expected_contents=[file_1_content],
        ),
    )
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    with TemporaryFile(file_1_content) as test_file:

        async with Client(transport=transport) as session:

            query = gql(file_upload_mutation_1)

            file_path = test_file.filename

            # Using an opened file
            with open(file_path, "rb") as f:

                query.variable_values = {"file": f, "other_var": 42}

                # Execute query asynchronously
                with pytest.warns(
                    DeprecationWarning,
                    match="Not using FileVar for file upload is deprecated",
                ):
                    result = await session.execute(query, upload_files=True)

            success = result["success"]
            assert success

            # Using an opened file inside a FileVar object
            with open(file_path, "rb") as f:

                query.variable_values = {"file": FileVar(f), "other_var": 42}

                with warnings.catch_warnings():
                    warnings.simplefilter("error")  # Turn warnings into errors
                    result = await session.execute(query, upload_files=True)

            success = result["success"]
            assert success

            # Using an filename string inside a FileVar object
            query.variable_values = {"file": FileVar(file_path), "other_var": 42}

            result = await session.execute(query, upload_files=True)

            success = result["success"]
            assert success


@pytest.mark.asyncio
async def test_aiohttp_file_upload_with_content_type(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    app = web.Application()
    app.router.add_route(
        "POST",
        "/",
        make_upload_handler(
            file_headers=[{"Content-Type": "application/pdf"}],
            expected_map=file_upload_mutation_1_map,
            expected_operations=file_upload_mutation_1_operations,
            expected_contents=[file_1_content],
        ),
    )
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    with TemporaryFile(file_1_content) as test_file:

        async with Client(transport=transport) as session:

            query = gql(file_upload_mutation_1)

            file_path = test_file.filename

            # Using an opened file
            with open(file_path, "rb") as f:

                # Setting the content_type
                f.content_type = "application/pdf"  # type: ignore

                query.variable_values = {"file": f, "other_var": 42}

                with pytest.warns(
                    DeprecationWarning,
                    match="Not using FileVar for file upload is deprecated",
                ):
                    result = await session.execute(query, upload_files=True)

            success = result["success"]
            assert success

            # Using an opened file inside a FileVar object
            with open(file_path, "rb") as f:

                query.variable_values = {
                    "file": FileVar(
                        f,
                        content_type="application/pdf",
                    ),
                    "other_var": 42,
                }

                result = await session.execute(query, upload_files=True)

            success = result["success"]
            assert success

            # Using an filename string inside a FileVar object
            query.variable_values = {
                "file": FileVar(
                    file_path,
                    content_type="application/pdf",
                ),
                "other_var": 42,
            }

            result = await session.execute(query, upload_files=True)

            success = result["success"]
            assert success


@pytest.mark.asyncio
async def test_aiohttp_file_upload_default_filename_is_basename(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    app = web.Application()

    with TemporaryFile(file_1_content) as test_file:
        file_path = test_file.filename
        file_basename = os.path.basename(file_path)

        app.router.add_route(
            "POST",
            "/",
            make_upload_handler(
                filenames=[file_basename],
                expected_map=file_upload_mutation_1_map,
                expected_operations=file_upload_mutation_1_operations,
                expected_contents=[file_1_content],
            ),
        )
        server = await aiohttp_server(app)

        url = server.make_url("/")

        transport = AIOHTTPTransport(url=url, timeout=10)

        async with Client(transport=transport) as session:

            query = gql(file_upload_mutation_1)

            query.variable_values = {
                "file": FileVar(
                    file_path,
                ),
                "other_var": 42,
            }

            result = await session.execute(query, upload_files=True)

            success = result["success"]
            assert success


@pytest.mark.asyncio
async def test_aiohttp_file_upload_with_filename(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    app = web.Application()

    with TemporaryFile(file_1_content) as test_file:
        file_path = test_file.filename

        app.router.add_route(
            "POST",
            "/",
            make_upload_handler(
                filenames=["filename1.txt"],
                expected_map=file_upload_mutation_1_map,
                expected_operations=file_upload_mutation_1_operations,
                expected_contents=[file_1_content],
            ),
        )
        server = await aiohttp_server(app)

        url = server.make_url("/")

        transport = AIOHTTPTransport(url=url, timeout=10)

        async with Client(transport=transport) as session:

            query = gql(file_upload_mutation_1)

            query.variable_values = {
                "file": FileVar(
                    file_path,
                    filename="filename1.txt",
                ),
                "other_var": 42,
            }

            result = await session.execute(query, upload_files=True)

            success = result["success"]
            assert success


@pytest.mark.asyncio
async def test_aiohttp_file_upload_without_session(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    app = web.Application()
    app.router.add_route(
        "POST",
        "/",
        make_upload_handler(
            expected_map=file_upload_mutation_1_map,
            expected_operations=file_upload_mutation_1_operations,
            expected_contents=[file_1_content],
        ),
    )
    server = await aiohttp_server(app)

    url = server.make_url("/")

    def test_code():
        transport = AIOHTTPTransport(url=url, timeout=10)

        with TemporaryFile(file_1_content) as test_file:

            client = Client(transport=transport)

            query = gql(file_upload_mutation_1)

            file_path = test_file.filename

            query.variable_values = {"file": FileVar(file_path), "other_var": 42}

            result = client.execute(query, upload_files=True)

            success = result["success"]
            assert success

    await run_sync_test(server, test_code)


@pytest.mark.asyncio
async def test_aiohttp_binary_file_upload(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    # This is a sample binary file content containing all possible byte values
    binary_file_content = bytes(range(0, 256))

    app = web.Application()
    app.router.add_route(
        "POST",
        "/",
        make_upload_handler(
            binary=True,
            expected_contents=[binary_file_content],
            expected_map=file_upload_mutation_1_map,
            expected_operations=file_upload_mutation_1_operations,
        ),
    )
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    with TemporaryFile(binary_file_content) as test_file:

        async with Client(transport=transport) as session:

            query = gql(file_upload_mutation_1)

            file_path = test_file.filename

            query.variable_values = {"file": FileVar(file_path), "other_var": 42}

            # Execute query asynchronously
            result = await session.execute(query, upload_files=True)

            success = result["success"]

            assert success


@pytest.mark.asyncio
async def test_aiohttp_stream_reader_upload(aiohttp_server):
    from aiohttp import ClientSession, web

    from gql.transport.aiohttp import AIOHTTPTransport

    # This is a sample binary file content containing all possible byte values
    binary_file_content = bytes(range(0, 256))

    async def binary_data_handler(request):
        return web.Response(
            body=binary_file_content, content_type="binary/octet-stream"
        )

    app = web.Application()
    app.router.add_route(
        "POST",
        "/",
        make_upload_handler(
            binary=True,
            expected_contents=[binary_file_content],
            expected_map=file_upload_mutation_1_map,
            expected_operations=file_upload_mutation_1_operations,
        ),
    )
    app.router.add_route("GET", "/binary_data", binary_data_handler)

    server = await aiohttp_server(app)

    url = server.make_url("/")
    binary_data_url = server.make_url("/binary_data")

    transport = AIOHTTPTransport(url=url, timeout=10)

    # Not using FileVar
    async with Client(transport=transport) as session:
        query = gql(file_upload_mutation_1)
        async with ClientSession() as client:
            async with client.get(binary_data_url) as resp:
                query.variable_values = {"file": resp.content, "other_var": 42}

                with pytest.warns(
                    DeprecationWarning,
                    match="Not using FileVar for file upload is deprecated",
                ):
                    result = await session.execute(query, upload_files=True)

    success = result["success"]
    assert success

    # Using FileVar
    async with Client(transport=transport) as session:
        query = gql(file_upload_mutation_1)
        async with ClientSession() as client:
            async with client.get(binary_data_url) as resp:
                query.variable_values = {"file": FileVar(resp.content), "other_var": 42}

                result = await session.execute(query, upload_files=True)

    success = result["success"]
    assert success


@pytest.mark.asyncio
async def test_aiohttp_async_generator_upload(aiohttp_server):
    import aiofiles
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    # This is a sample binary file content containing all possible byte values
    binary_file_content = bytes(range(0, 256))

    app = web.Application()
    app.router.add_route(
        "POST",
        "/",
        make_upload_handler(
            binary=True,
            expected_contents=[binary_file_content],
            expected_map=file_upload_mutation_1_map,
            expected_operations=file_upload_mutation_1_operations,
        ),
    )
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    query = gql(file_upload_mutation_1)

    with TemporaryFile(binary_file_content) as test_file:

        file_path = test_file.filename

        async def file_sender(file_name):
            async with aiofiles.open(file_name, "rb") as f:
                chunk = await f.read(64 * 1024)
                while chunk:
                    yield chunk
                    chunk = await f.read(64 * 1024)

        # Not using FileVar
        async with Client(transport=transport) as session:

            query.variable_values = {"file": file_sender(file_path), "other_var": 42}

            with pytest.warns(
                DeprecationWarning,
                match="Not using FileVar for file upload is deprecated",
            ):
                result = await session.execute(query, upload_files=True)

            success = result["success"]
            assert success

        # Using FileVar
        async with Client(transport=transport) as session:

            query.variable_values = {
                "file": FileVar(file_sender(file_path)),
                "other_var": 42,
            }

            # Execute query asynchronously
            result = await session.execute(query, upload_files=True)

            success = result["success"]
            assert success

        # Using FileVar with new streaming support
        async with Client(transport=transport) as session:

            query.variable_values = {
                "file": FileVar(file_path, streaming=True),
                "other_var": 42,
            }

            # Execute query asynchronously
            result = await session.execute(query, upload_files=True)

            success = result["success"]
            assert success


@pytest.mark.asyncio
async def test_aiohttp_file_upload_two_files(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    file_upload_mutation_2 = """
    mutation($file1: Upload!, $file2: Upload!) {
      uploadFile(input:{file1:$file, file2:$file}) {
        success
      }
    }
    """

    file_upload_mutation_2_operations = (
        '{"query": "mutation ($file1: Upload!, $file2: Upload!) {\\n  '
        'uploadFile(input: {file1: $file, file2: $file}) {\\n    success\\n  }\\n}", '
        '"variables": {"file1": null, "file2": null}}'
    )

    file_upload_mutation_2_map = '{"0": ["variables.file1"], "1": ["variables.file2"]}'

    file_2_content = """
    This is a second test file
    This file will also be sent in the GraphQL mutation
    """

    app = web.Application()
    app.router.add_route(
        "POST",
        "/",
        make_upload_handler(
            nb_files=2,
            expected_map=file_upload_mutation_2_map,
            expected_operations=file_upload_mutation_2_operations,
            expected_contents=[file_1_content, file_2_content],
        ),
    )
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    with TemporaryFile(file_1_content) as test_file_1:
        with TemporaryFile(file_2_content) as test_file_2:

            async with Client(transport=transport) as session:

                query = gql(file_upload_mutation_2)

                file_path_1 = test_file_1.filename
                file_path_2 = test_file_2.filename

                query.variable_values = {
                    "file1": FileVar(file_path_1),
                    "file2": FileVar(file_path_2),
                }

                result = await session.execute(query, upload_files=True)

                success = result["success"]

                assert success


@pytest.mark.asyncio
async def test_aiohttp_file_upload_list_of_two_files(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    file_upload_mutation_3 = """
    mutation($files: [Upload!]!) {
      uploadFiles(input:{files:$files}) {
        success
      }
    }
    """

    file_upload_mutation_3_operations = (
        '{"query": "mutation ($files: [Upload!]!) {\\n  uploadFiles'
        "(input: {files: $files})"
        ' {\\n    success\\n  }\\n}", "variables": {"files": [null, null]}}'
    )

    file_upload_mutation_3_map = (
        '{"0": ["variables.files.0"], "1": ["variables.files.1"]}'
    )

    file_2_content = """
    This is a second test file
    This file will also be sent in the GraphQL mutation
    """

    app = web.Application()
    app.router.add_route(
        "POST",
        "/",
        make_upload_handler(
            nb_files=2,
            expected_map=file_upload_mutation_3_map,
            expected_operations=file_upload_mutation_3_operations,
            expected_contents=[file_1_content, file_2_content],
        ),
    )
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    with TemporaryFile(file_1_content) as test_file_1:
        with TemporaryFile(file_2_content) as test_file_2:

            async with Client(transport=transport) as session:

                query = gql(file_upload_mutation_3)

                file_path_1 = test_file_1.filename
                file_path_2 = test_file_2.filename

                query.variable_values = {
                    "files": [
                        FileVar(file_path_1),
                        FileVar(file_path_2),
                    ],
                }

                # Execute query asynchronously
                result = await session.execute(query, upload_files=True)

                success = result["success"]

                assert success


@pytest.mark.asyncio
async def test_aiohttp_using_cli(aiohttp_server, monkeypatch, capsys):
    from aiohttp import web

    async def handler(request):
        return web.Response(text=query1_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    parser = get_parser(with_examples=True)
    args = parser.parse_args([url, "--verbose"])

    # Monkeypatching sys.stdin to simulate getting the query
    # via the standard input
    monkeypatch.setattr("sys.stdin", io.StringIO(query1_str))

    exit_code = await main(args)

    assert exit_code == 0

    # Check that the result has been printed on stdout
    captured = capsys.readouterr()
    captured_out = str(captured.out).strip()

    expected_answer = json.loads(query1_server_answer_data)
    print(f"Captured: {captured_out}")
    received_answer = json.loads(captured_out)

    assert received_answer == expected_answer


@pytest.mark.asyncio
@pytest.mark.script_launch_mode("subprocess")
async def test_aiohttp_using_cli_ep(
    aiohttp_server, monkeypatch, script_runner, run_sync_test
):
    from aiohttp import web

    async def handler(request):
        return web.Response(text=query1_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    def test_code():

        monkeypatch.setattr("sys.stdin", io.StringIO(query1_str))

        ret = script_runner.run(
            ["gql-cli", url, "--verbose"],
            stdin=io.StringIO(query1_str),
        )

        assert ret.success

        # Check that the result has been printed on stdout
        captured_out = str(ret.stdout).strip()

        expected_answer = json.loads(query1_server_answer_data)
        print(f"Captured: {captured_out}")
        received_answer = json.loads(captured_out)

        assert received_answer == expected_answer

    await run_sync_test(server, test_code)


@pytest.mark.asyncio
async def test_aiohttp_using_cli_invalid_param(aiohttp_server, monkeypatch, capsys):
    from aiohttp import web

    async def handler(request):
        return web.Response(text=query1_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    parser = get_parser(with_examples=True)
    args = parser.parse_args([url, "--variables", "invalid_param"])

    # Monkeypatching sys.stdin to simulate getting the query
    # via the standard input
    monkeypatch.setattr("sys.stdin", io.StringIO(query1_str))

    # Check that the exit_code is an error
    exit_code = await main(args)
    assert exit_code == 1

    # Check that the error has been printed on stdout
    captured = capsys.readouterr()
    captured_err = str(captured.err).strip()
    print(f"Captured: {captured_err}")

    expected_error = "Error: Invalid variable: invalid_param"

    assert expected_error in captured_err


@pytest.mark.asyncio
async def test_aiohttp_using_cli_invalid_query(aiohttp_server, monkeypatch, capsys):
    from aiohttp import web

    async def handler(request):
        return web.Response(text=query1_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    parser = get_parser(with_examples=True)
    args = parser.parse_args([url])

    # Send invalid query on standard input
    monkeypatch.setattr("sys.stdin", io.StringIO("BLAHBLAH"))

    exit_code = await main(args)

    assert exit_code == 1

    # Check that the error has been printed on stdout
    captured = capsys.readouterr()
    captured_err = str(captured.err).strip()
    print(f"Captured: {captured_err}")

    expected_error = "Syntax Error: Unexpected Name 'BLAHBLAH'"

    assert expected_error in captured_err


query1_server_answer_with_extensions = (
    f'{{"data":{query1_server_answer_data}, "extensions":{{"key1": "val1"}}}}'
)


@pytest.mark.asyncio
async def test_aiohttp_query_with_extensions(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(
            text=query1_server_answer_with_extensions, content_type="application/json"
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    async with Client(transport=transport) as session:

        query = gql(query1_str)

        execution_result = await session.execute(query, get_execution_result=True)

        assert execution_result.extensions["key1"] == "val1"


@pytest.mark.asyncio
@pytest.mark.parametrize("ssl_close_timeout", [0, 10])
@pytest.mark.parametrize("verify_https", ["disabled", "cert_provided"])
async def test_aiohttp_query_https(ssl_aiohttp_server, ssl_close_timeout, verify_https):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(text=query1_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await ssl_aiohttp_server(app)

    url = server.make_url("/")

    assert str(url).startswith("https://")

    extra_args = {}

    if verify_https == "cert_provided":
        _, ssl_context = get_localhost_ssl_context_client()

        extra_args["ssl"] = ssl_context
    elif verify_https == "disabled":
        extra_args["ssl"] = False

    transport = AIOHTTPTransport(
        url=url,
        timeout=10,
        ssl_close_timeout=ssl_close_timeout,
        **extra_args,
    )

    async with Client(transport=transport) as session:

        query = gql(query1_str)

        # Execute query asynchronously
        result = await session.execute(query)

        continents = result["continents"]

        africa = continents[0]

        assert africa["code"] == "AF"


@pytest.mark.asyncio
async def test_aiohttp_query_https_self_cert_fail(ssl_aiohttp_server):
    """By default, we should verify the ssl certificate"""
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(text=query1_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await ssl_aiohttp_server(app)

    url = server.make_url("/")

    assert str(url).startswith("https://")

    transport = AIOHTTPTransport(url=url, timeout=10)

    query = gql(query1_str)

    expected_error = "certificate verify failed: self-signed certificate"

    with pytest.raises(TransportConnectionFailed) as exc_info:
        async with Client(transport=transport) as session:
            await session.execute(query)

    assert expected_error in str(exc_info.value)

    with pytest.raises(TransportConnectionFailed) as exc_info:
        async with Client(transport=transport) as session:
            await session.execute_batch([query])

    assert expected_error in str(exc_info.value)

    assert transport.session is None


@pytest.mark.asyncio
async def test_aiohttp_query_https_self_cert_default(ssl_aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(text=query1_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await ssl_aiohttp_server(app)

    url = server.make_url("/")

    assert str(url).startswith("https://")

    transport = AIOHTTPTransport(url=url)

    assert transport.ssl is True


@pytest.mark.asyncio
async def test_aiohttp_error_fetching_schema(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    error_answer = """
{
    "errors": [
        {
            "errorType": "UnauthorizedException",
            "message": "Permission denied"
        }
    ]
}
"""

    async def handler(request):
        return web.Response(
            text=error_answer,
            content_type="application/json",
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    with pytest.raises(TransportQueryError) as exc_info:
        async with Client(transport=transport, fetch_schema_from_transport=True):
            pass

    expected_error = (
        "Error while fetching schema: "
        "{'errorType': 'UnauthorizedException', 'message': 'Permission denied'}"
    )

    assert expected_error in str(exc_info.value)
    assert transport.session is None


@pytest.mark.asyncio
async def test_aiohttp_reconnecting_session(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(
            text=query1_server_answer,
            content_type="application/json",
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    client = Client(transport=transport)

    session = await client.connect_async(reconnecting=True)

    query = gql(query1_str)

    # Execute query asynchronously
    result = await session.execute(query)

    continents = result["continents"]

    africa = continents[0]

    assert africa["code"] == "AF"

    await client.close_async()


@pytest.mark.asyncio
@pytest.mark.parametrize("retries", [False, lambda e: e])
async def test_aiohttp_reconnecting_session_retries(aiohttp_server, retries):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(
            text=query1_server_answer,
            content_type="application/json",
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    client = Client(transport=transport)

    session = await client.connect_async(
        reconnecting=True, retry_execute=retries, retry_connect=retries
    )

    assert session._execute_with_retries == session._execute_once
    assert session._connect_with_retries == session.transport.connect

    await client.close_async()


@pytest.mark.asyncio
async def test_aiohttp_reconnecting_session_start_connecting_task_twice(
    aiohttp_server, caplog
):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(
            text=query1_server_answer,
            content_type="application/json",
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    client = Client(transport=transport)

    session = await client.connect_async(reconnecting=True)

    await session.start_connecting_task()

    print(f"Captured log: {caplog.text}")

    expected_warning = "connect task already started!"
    assert expected_warning in caplog.text

    await client.close_async()


@pytest.mark.asyncio
async def test_aiohttp_json_serializer(aiohttp_server, caplog):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):

        request_text = await request.text()
        print("Received on backend: " + request_text)

        return web.Response(
            text=query1_server_answer,
            content_type="application/json",
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(
        url=url,
        timeout=10,
        json_serialize=lambda e: json.dumps(e, separators=(",", ":")),
    )

    async with Client(transport=transport) as session:

        query = gql(query1_str)

        # Execute query asynchronously
        result = await session.execute(query)

        continents = result["continents"]

        africa = continents[0]

        assert africa["code"] == "AF"

    # Checking that there is no space after the colon in the log
    expected_log = '"query":"query getContinents'
    assert expected_log in caplog.text


query_float_str = """
    query getPi {
      pi
    }
"""

query_float_server_answer_data = '{"pi": 3.141592653589793238462643383279502884197}'

query_float_server_answer = f'{{"data":{query_float_server_answer_data}}}'


@pytest.mark.asyncio
async def test_aiohttp_json_deserializer(aiohttp_server):
    from decimal import Decimal
    from functools import partial

    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(
            text=query_float_server_answer,
            content_type="application/json",
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    json_loads = partial(json.loads, parse_float=Decimal)

    transport = AIOHTTPTransport(
        url=url,
        timeout=10,
        json_deserialize=json_loads,
    )

    async with Client(transport=transport) as session:

        query = gql(query_float_str)

        # Execute query asynchronously
        result = await session.execute(query)

        pi = result["pi"]

        assert pi == Decimal("3.141592653589793238462643383279502884197")


@pytest.mark.asyncio
async def test_aiohttp_connector_owner_false(aiohttp_server):
    from aiohttp import TCPConnector, web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(
            text=query1_server_answer,
            content_type="application/json",
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    connector = TCPConnector()
    transport = AIOHTTPTransport(
        url=url,
        timeout=10,
        client_session_args={
            "connector": connector,
            "connector_owner": False,
        },
    )

    for _ in range(2):
        async with Client(transport=transport) as session:

            query = gql(query1_str)

            # Execute query asynchronously
            result = await session.execute(query)

            continents = result["continents"]

            africa = continents[0]

            assert africa["code"] == "AF"

    await connector.close()


@pytest.mark.asyncio
async def test_aiohttp_deprecation_warning_using_document_node_execute(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(
            text=query1_server_answer,
            content_type="application/json",
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    async with Client(transport=transport) as session:

        query = gql(query1_str)

        with pytest.warns(
            DeprecationWarning,
            match="Using a DocumentNode is deprecated",
        ):
            result = await session.execute(query.document)

        continents = result["continents"]

        africa = continents[0]

        assert africa["code"] == "AF"


@pytest.mark.asyncio
async def test_aiohttp_deprecation_warning_execute_variable_values(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(text=query2_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    async with Client(transport=transport) as session:

        query = gql(query2_str)

        with pytest.warns(
            DeprecationWarning,
            match=(
                "Using variable_values and operation_name arguments of "
                "execute and subscribe methods is deprecated"
            ),
        ):
            result = await session.execute(
                query,
                variable_values={"code": "EU"},
                operation_name="getEurope",
            )

        continent = result["continent"]

        assert continent["name"] == "Europe"


@pytest.mark.asyncio
async def test_aiohttp_type_error_execute(aiohttp_server):
    from aiohttp import web

    from gql.transport.aiohttp import AIOHTTPTransport

    async def handler(request):
        return web.Response(text=query2_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    async with Client(transport=transport) as session:

        with pytest.raises(TypeError) as exc_info:
            await session.execute("qmlsdkfj")

        assert "request should be a GraphQLRequest object" in str(exc_info.value)



import io
import os
import warnings
from typing import Any, Dict, List, Optional, Tuple, Type


class FileVar:
    def __init__(
        self,
        f: Any,  # str | io.IOBase | aiohttp.StreamReader | AsyncGenerator
        *,
        filename: Optional[str] = None,
        content_type: Optional[str] = None,
        streaming: bool = False,
        streaming_block_size: int = 64 * 1024,
    ):
        self.f = f
        self.filename = filename
        self.content_type = content_type
        self.streaming = streaming
        self.streaming_block_size = streaming_block_size

        self._file_opened: bool = False

    def open_file(
        self,
        transport_supports_streaming: bool = False,
    ) -> None:
        assert self._file_opened is False

        if self.streaming:
            assert (
                transport_supports_streaming
            ), "streaming not supported on this transport"
            self._make_file_streamer()
        else:
            if isinstance(self.f, str):
                if self.filename is None:
                    # By default we set the filename to the basename
                    # of the opened file
                    self.filename = os.path.basename(self.f)
                self.f = open(self.f, "rb")
                self._file_opened = True

    def close_file(self) -> None:
        if self._file_opened:
            assert isinstance(self.f, io.IOBase)
            self.f.close()
            self._file_opened = False

    def _make_file_streamer(self) -> None:
        assert isinstance(self.f, str), "streaming option needs a filepath str"

        import aiofiles

        async def file_sender(file_name):
            async with aiofiles.open(file_name, "rb") as f:
                while chunk := await f.read(self.streaming_block_size):
                    yield chunk

        self.f = file_sender(self.f)


def open_files(
    filevars: List[FileVar],
    transport_supports_streaming: bool = False,
) -> None:

    for filevar in filevars:
        filevar.open_file(transport_supports_streaming=transport_supports_streaming)


def close_files(filevars: List[FileVar]) -> None:
    for filevar in filevars:
        filevar.close_file()


FILE_UPLOAD_DOCS = "https://gql.readthedocs.io/en/latest/usage/file_upload.html"


def extract_files(
    variables: Dict, file_classes: Tuple[Type[Any], ...]
) -> Tuple[Dict, Dict[str, FileVar]]:
    files: Dict[str, FileVar] = {}

    def recurse_extract(path, obj):
        """
        recursively traverse obj, doing a deepcopy, but
        replacing any file-like objects with nulls and
        shunting the originals off to the side.
        """
        nonlocal files
        if isinstance(obj, list):
            nulled_list = []
            for key, value in enumerate(obj):
                value = recurse_extract(f"{path}.{key}", value)
                nulled_list.append(value)
            return nulled_list
        elif isinstance(obj, dict):
            nulled_dict = {}
            for key, value in obj.items():
                value = recurse_extract(f"{path}.{key}", value)
                nulled_dict[key] = value
            return nulled_dict
        elif isinstance(obj, file_classes):
            # extract obj from its parent and put it into files instead.
            warnings.warn(
                "Not using FileVar for file upload is deprecated. "
                f"See {FILE_UPLOAD_DOCS} for details.",
                DeprecationWarning,
            )
            name = getattr(obj, "name", None)
            content_type = getattr(obj, "content_type", None)
            files[path] = FileVar(obj, filename=name, content_type=content_type)
            return None
        elif isinstance(obj, FileVar):
            # extract obj from its parent and put it into files instead.
            files[path] = obj
            return None
        else:
            # base case: pass through unchanged
            return obj

    nulled_variables = recurse_extract("variables", variables)

    return nulled_variables, files



from ssl import SSLContext
from typing import Any, Dict, List, Literal, Mapping, Optional, Union

from aiohttp import BasicAuth, ClientSession, Fingerprint
from aiohttp.typedefs import LooseHeaders, StrOrURL

from .common.adapters.aiohttp import AIOHTTPWebSocketsAdapter
from .websockets_protocol import WebsocketsProtocolTransportBase


class AIOHTTPWebsocketsTransport(WebsocketsProtocolTransportBase):
    """:ref:`Async Transport <async_transports>` used to execute GraphQL queries on
    remote servers with websocket connection.

    This transport uses asyncio and the provided aiohttp adapter library
    in order to send requests on a websocket connection.
    """

    def __init__(
        self,
        url: StrOrURL,
        *,
        subprotocols: Optional[List[str]] = None,
        heartbeat: Optional[float] = None,
        auth: Optional[BasicAuth] = None,
        origin: Optional[str] = None,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[LooseHeaders] = None,
        proxy: Optional[StrOrURL] = None,
        proxy_auth: Optional[BasicAuth] = None,
        proxy_headers: Optional[LooseHeaders] = None,
        ssl: Optional[Union[SSLContext, Literal[False], Fingerprint]] = None,
        websocket_close_timeout: float = 10.0,
        receive_timeout: Optional[float] = None,
        ssl_close_timeout: Optional[Union[int, float]] = 10,
        connect_timeout: Optional[Union[int, float]] = 10,
        close_timeout: Optional[Union[int, float]] = 10,
        ack_timeout: Optional[Union[int, float]] = 10,
        keep_alive_timeout: Optional[Union[int, float]] = None,
        init_payload: Dict[str, Any] = {},
        ping_interval: Optional[Union[int, float]] = None,
        pong_timeout: Optional[Union[int, float]] = None,
        answer_pings: bool = True,
        session: Optional[ClientSession] = None,
        client_session_args: Optional[Dict[str, Any]] = None,
        connect_args: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Initialize the transport with the given parameters.

        :param url: The GraphQL server URL. Example: 'wss://server.com:PORT/graphql'.
        :param subprotocols: list of subprotocols sent to the
            backend in the 'subprotocols' http header.
            By default: both apollo and graphql-ws subprotocols.
        :param float heartbeat: Send low level `ping` message every `heartbeat`
                                seconds and wait `pong` response, close
                                connection if `pong` response is not
                                received. The timer is reset on any data reception.
        :param auth: An object that represents HTTP Basic Authorization.
                     :class:`~aiohttp.BasicAuth` (optional)
        :param str origin: Origin header to send to server(optional)
        :param params: Mapping, iterable of tuple of *key*/*value* pairs or
                       string to be sent as parameters in the query
                       string of the new request. Ignored for subsequent
                       redirected requests (optional)

                       Allowed values are:

                       - :class:`collections.abc.Mapping` e.g. :class:`dict`,
                         :class:`multidict.MultiDict` or
                         :class:`multidict.MultiDictProxy`
                       - :class:`collections.abc.Iterable` e.g. :class:`tuple` or
                         :class:`list`
                       - :class:`str` with preferably url-encoded content
                         (**Warning:** content will not be encoded by *aiohttp*)
        :param headers: HTTP Headers that sent with every request
                        May be either *iterable of key-value pairs* or
                        :class:`~collections.abc.Mapping`
                        (e.g. :class:`dict`,
                        :class:`~multidict.CIMultiDict`).
        :param proxy: Proxy URL, :class:`str` or :class:`~yarl.URL` (optional)
        :param aiohttp.BasicAuth proxy_auth: an object that represents proxy HTTP
                                             Basic Authorization (optional)
        :param ssl: SSL validation mode. ``True`` for default SSL check
                      (:func:`ssl.create_default_context` is used),
                      ``False`` for skip SSL certificate validation,
                      :class:`aiohttp.Fingerprint` for fingerprint
                      validation, :class:`ssl.SSLContext` for custom SSL
                      certificate validation.
        :param float websocket_close_timeout: Timeout for websocket to close.
                                              ``10`` seconds by default
        :param float receive_timeout: Timeout for websocket to receive
                                      complete message.  ``None`` (unlimited)
                                      seconds by default
        :param ssl_close_timeout: Timeout in seconds to wait for the ssl connection
                                  to close properly
        :param connect_timeout: Timeout in seconds for the establishment
            of the websocket connection. If None is provided this will wait forever.
        :param close_timeout: Timeout in seconds for the close. If None is provided
            this will wait forever.
        :param ack_timeout: Timeout in seconds to wait for the connection_ack message
            from the server. If None is provided this will wait forever.
        :param keep_alive_timeout: Optional Timeout in seconds to receive
            a sign of liveness from the server.
        :param init_payload: Dict of the payload sent in the connection_init message.
        :param ping_interval: Delay in seconds between pings sent by the client to
            the backend for the graphql-ws protocol. None (by default) means that
            we don't send pings. Note: there are also pings sent by the underlying
            websockets protocol. See the
            :ref:`keepalive documentation <websockets_transport_keepalives>`
            for more information about this.
        :param pong_timeout: Delay in seconds to receive a pong from the backend
            after we sent a ping (only for the graphql-ws protocol).
            By default equal to half of the ping_interval.
        :param answer_pings: Whether the client answers the pings from the backend
            (for the graphql-ws protocol).
            By default: True
        :param session: Optional aiohttp.ClientSession instance.
        :param client_session_args: Dict of extra args passed to
                `aiohttp.ClientSession`_
        :param connect_args: Dict of extra args passed to
                `aiohttp.ClientSession.ws_connect`_

        .. _aiohttp.ClientSession.ws_connect:
          https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.ClientSession.ws_connect
        .. _aiohttp.ClientSession:
          https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.ClientSession
        """

        # Instanciate a AIOHTTPWebSocketAdapter to indicate the use
        # of the aiohttp dependency for this transport
        self.adapter: AIOHTTPWebSocketsAdapter = AIOHTTPWebSocketsAdapter(
            url=url,
            headers=headers,
            ssl=ssl,
            session=session,
            client_session_args=client_session_args,
            connect_args=connect_args,
            heartbeat=heartbeat,
            auth=auth,
            origin=origin,
            params=params,
            proxy=proxy,
            proxy_auth=proxy_auth,
            proxy_headers=proxy_headers,
            websocket_close_timeout=websocket_close_timeout,
            receive_timeout=receive_timeout,
            ssl_close_timeout=ssl_close_timeout,
        )

        # Initialize the WebsocketsProtocolTransportBase parent class
        super().__init__(
            adapter=self.adapter,
            init_payload=init_payload,
            connect_timeout=connect_timeout,
            close_timeout=close_timeout,
            ack_timeout=ack_timeout,
            keep_alive_timeout=keep_alive_timeout,
            ping_interval=ping_interval,
            pong_timeout=pong_timeout,
            answer_pings=answer_pings,
            subprotocols=subprotocols,
        )

    @property
    def headers(self) -> Optional[LooseHeaders]:
        return self.adapter.headers

    @property
    def ssl(self) -> Optional[Union[SSLContext, Literal[False], Fingerprint]]:
        return self.adapter.ssl
